/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
struct toto {
   
};
int main(void)
{
  int __retres;
  struct toto s;
  struct toto *p;
  __e_acsl_memory_init((int *)0,(char ***)0,(size_t)8);
  __e_acsl_store_block((void *)(& p),(size_t)8);
  __e_acsl_store_block((void *)(& s),(size_t)0);
  {
    int __gen_e_acsl_valid;
    __gen_e_acsl_valid = __e_acsl_valid((void *)(& s),sizeof(struct toto),
                                        (void *)(& s),(void *)0);
    __e_acsl_assert(__gen_e_acsl_valid,"Assertion","main","\\valid(&s)",
                    "tests/bts/bts1700.i",9);
  }
  /*@ assert \valid(&s); */ ;
  __e_acsl_full_init((void *)(& p));
  p = & s;
  {
    int __gen_e_acsl_initialized;
    int __gen_e_acsl_and;
    __gen_e_acsl_initialized = __e_acsl_initialized((void *)(& p),
                                                    sizeof(struct toto *));
    if (__gen_e_acsl_initialized) {
      int __gen_e_acsl_valid_2;
      __gen_e_acsl_valid_2 = __e_acsl_valid((void *)p,sizeof(struct toto),
                                            (void *)p,(void *)(& p));
      __gen_e_acsl_and = __gen_e_acsl_valid_2;
    }
    else __gen_e_acsl_and = 0;
    __e_acsl_assert(__gen_e_acsl_and,"Assertion","main","\\valid(p)",
                    "tests/bts/bts1700.i",12);
  }
  /*@ assert \valid(p); */ ;
  __retres = 0;
  __e_acsl_delete_block((void *)(& p));
  __e_acsl_delete_block((void *)(& s));
  __e_acsl_memory_clean();
  return __retres;
}


