/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
int a = 0;
int b;
void __e_acsl_globals_init(void)
{
  static char __e_acsl_already_run = 0;
  if (! __e_acsl_already_run) {
    __e_acsl_already_run = 1;
    __e_acsl_store_block((void *)(& b),(size_t)4);
    __e_acsl_full_init((void *)(& b));
    __e_acsl_store_block((void *)(& a),(size_t)4);
    __e_acsl_full_init((void *)(& a));
  }
  return;
}

void __e_acsl_globals_clean(void)
{
  __e_acsl_delete_block((void *)(& b));
  __e_acsl_delete_block((void *)(& a));
}

int main(void)
{
  int __retres;
  __e_acsl_memory_init((int *)0,(char ***)0,(size_t)8);
  __e_acsl_globals_init();
  int *p = & a;
  __e_acsl_store_block((void *)(& p),(size_t)8);
  __e_acsl_full_init((void *)(& p));
  int *q = & b;
  __e_acsl_store_block((void *)(& q),(size_t)8);
  __e_acsl_full_init((void *)(& q));
  __e_acsl_assert(1,"Assertion","main","\\initialized(&b)",
                  "tests/memory/init.c",9);
  /*@ assert \initialized(&b); */ ;
  {
    int __gen_e_acsl_initialized;
    __gen_e_acsl_initialized = __e_acsl_initialized((void *)q,sizeof(int));
    __e_acsl_assert(__gen_e_acsl_initialized,"Assertion","main",
                    "\\initialized(q)","tests/memory/init.c",10);
  }
  /*@ assert \initialized(q); */ ;
  {
    int __gen_e_acsl_initialized_2;
    __gen_e_acsl_initialized_2 = __e_acsl_initialized((void *)p,sizeof(int));
    __e_acsl_assert(__gen_e_acsl_initialized_2,"Assertion","main",
                    "\\initialized(p)","tests/memory/init.c",11);
  }
  /*@ assert \initialized(p); */ ;
  __retres = 0;
  __e_acsl_delete_block((void *)(& q));
  __e_acsl_delete_block((void *)(& p));
  __e_acsl_globals_clean();
  __e_acsl_memory_clean();
  return __retres;
}


