/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
#include "stdlib.h"
char *__gen_e_acsl_literal_string;
extern int __e_acsl_sound_verdict;

/*@ requires valid_nptr: \valid_read(nptr);
    assigns \result;
    assigns \result \from (indirect: nptr), (indirect: *(nptr + (0 ..)));
 */
int __gen_e_acsl_atoi(char const *nptr);

int a;
char *n = (char *)"134";
/*@ requires valid_nptr: \valid_read(nptr);
    assigns \result;
    assigns \result \from (indirect: nptr), (indirect: *(nptr + (0 ..)));
 */
int __gen_e_acsl_atoi(char const *nptr)
{
  int __retres;
  {
    int __gen_e_acsl_valid_read;
    __e_acsl_store_block((void *)(& nptr),(size_t)8);
    __gen_e_acsl_valid_read = __e_acsl_valid_read((void *)nptr,
                                                  sizeof(char const),
                                                  (void *)nptr,
                                                  (void *)(& nptr));
    __e_acsl_assert(__gen_e_acsl_valid_read,"Precondition","atoi",
                    "\\valid_read(nptr)","FRAMAC_SHARE/libc/stdlib.h",78);
  }
  __retres = atoi(nptr);
  __e_acsl_delete_block((void *)(& nptr));
  return __retres;
}

void __e_acsl_globals_init(void)
{
  static char __e_acsl_already_run = 0;
  if (! __e_acsl_already_run) {
    __e_acsl_already_run = 1;
    __gen_e_acsl_literal_string = "134";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string,sizeof("134"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string);
    __e_acsl_store_block((void *)(& n),(size_t)8);
    __e_acsl_full_init((void *)(& n));
  }
  return;
}

void __e_acsl_globals_clean(void)
{
  __e_acsl_delete_block((void *)(& n));
}

int main(int argc, char **argv)
{
  int __retres;
  __e_acsl_memory_init(& argc,& argv,(size_t)8);
  __e_acsl_globals_init();
  argc = __gen_e_acsl_atoi((char const *)n);
  a = argc;
  __retres = 0;
  __e_acsl_globals_clean();
  __e_acsl_memory_clean();
  return __retres;
}


