/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
int main(void)
{
  int __retres;
  int x = 0;
  __e_acsl_assert(x == 0,"Assertion","main","x == 0","tests/arith/not.i",6);
  /*@ assert x ≡ 0; */ ;
  if (x) {
    __e_acsl_assert(x != 0,"Assertion","main","x != 0","tests/arith/not.i",7);
    /*@ assert x ≢ 0; */ ;
  }
  __retres = 0;
  return __retres;
}


