/* run.config_ci, run.config_dev
  COMMENT: Compile RTL with debug and debug verbose informations
  STDOPT:#"-e-acsl-debug 1"
  MACRO: ROOT_EACSL_GCC_OPTS_EXT --rt-debug --rt-verbose
 */
int main() {
  return 0;
}
