/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
char *__gen_e_acsl_literal_string;
void f(void const *s, int c, unsigned long n)
{
  __e_acsl_store_block((void *)(& s),(size_t)8);
  unsigned char const *p = (unsigned char const *)s;
  __e_acsl_store_block((void *)(& p),(size_t)8);
  __e_acsl_full_init((void *)(& p));
  {
    __e_acsl_mpz_t __gen_e_acsl_;
    __e_acsl_mpz_t __gen_e_acsl_n;
    __e_acsl_mpz_t __gen_e_acsl_sub;
    int __gen_e_acsl_eq;
    __gmpz_init_set_ui(__gen_e_acsl_,p - (unsigned char const *)s);
    __gmpz_init_set_ui(__gen_e_acsl_n,n);
    __gmpz_init(__gen_e_acsl_sub);
    __gmpz_sub(__gen_e_acsl_sub,
               (__e_acsl_mpz_struct const *)(__gen_e_acsl_n),
               (__e_acsl_mpz_struct const *)(__gen_e_acsl_n));
    __gen_e_acsl_eq = __gmpz_cmp((__e_acsl_mpz_struct const *)(__gen_e_acsl_),
                                 (__e_acsl_mpz_struct const *)(__gen_e_acsl_sub));
    __e_acsl_assert(__gen_e_acsl_eq == 0,"Assertion","f",
                    "p - (unsigned char const *)s == n - n",
                    "tests/bts/bts2386.c",7);
    __gmpz_clear(__gen_e_acsl_);
    __gmpz_clear(__gen_e_acsl_n);
    __gmpz_clear(__gen_e_acsl_sub);
  }
  /*@ assert p - (unsigned char const *)s ≡ n - n; */ ;
  __e_acsl_assert(p - (unsigned char const *)s == 0UL,"Assertion","f",
                  "p - (unsigned char const *)s == 0","tests/bts/bts2386.c",
                  8);
  /*@ assert p - (unsigned char const *)s ≡ 0; */ ;
  __e_acsl_delete_block((void *)(& s));
  __e_acsl_delete_block((void *)(& p));
  return;
}

void __e_acsl_globals_init(void)
{
  static char __e_acsl_already_run = 0;
  if (! __e_acsl_already_run) {
    __e_acsl_already_run = 1;
    __gen_e_acsl_literal_string = "1234567890";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string,
                         sizeof("1234567890"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string);
  }
  return;
}

int main(void)
{
  int __retres;
  __e_acsl_memory_init((int *)0,(char ***)0,(size_t)8);
  __e_acsl_globals_init();
  char const *s = __gen_e_acsl_literal_string;
  __e_acsl_store_block((void *)(& s),(size_t)8);
  __e_acsl_full_init((void *)(& s));
  f((void const *)s,'0',(unsigned long)11);
  __retres = 0;
  __e_acsl_delete_block((void *)(& s));
  __e_acsl_memory_clean();
  return __retres;
}


