/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
int main(void)
{
  int __retres;
  __e_acsl_assert((void *)0 == (void *)0,"Assertion","main",
                  "\\null == (void *)0","tests/memory/null.i",6);
  /*@ assert \null ≡ (void *)0; */ ;
  __retres = 0;
  return __retres;
}


