/* Generated by Frama-C */
#include "signal.h"
#include "stddef.h"
#include "stdio.h"
#include "stdlib.h"
#include "sys/select.h"
#include "sys/time.h"
#include "sys/types.h"
#include "sys/wait.h"
#include "unistd.h"
char *__gen_e_acsl_literal_string_31;
char *__gen_e_acsl_literal_string_30;
char *__gen_e_acsl_literal_string_29;
char *__gen_e_acsl_literal_string_28;
char *__gen_e_acsl_literal_string_27;
char *__gen_e_acsl_literal_string_26;
char *__gen_e_acsl_literal_string_25;
char *__gen_e_acsl_literal_string_24;
char *__gen_e_acsl_literal_string_23;
char *__gen_e_acsl_literal_string_21;
char *__gen_e_acsl_literal_string_19;
char *__gen_e_acsl_literal_string_16;
char *__gen_e_acsl_literal_string_15;
char *__gen_e_acsl_literal_string_14;
char *__gen_e_acsl_literal_string_13;
char *__gen_e_acsl_literal_string_12;
char *__gen_e_acsl_literal_string_9;
char *__gen_e_acsl_literal_string_8;
char *__gen_e_acsl_literal_string_11;
char *__gen_e_acsl_literal_string_7;
char *__gen_e_acsl_literal_string_10;
char *__gen_e_acsl_literal_string;
char *__gen_e_acsl_literal_string_2;
char *__gen_e_acsl_literal_string_3;
char *__gen_e_acsl_literal_string_6;
char *__gen_e_acsl_literal_string_5;
char *__gen_e_acsl_literal_string_4;
char *__gen_e_acsl_literal_string_22;
char *__gen_e_acsl_literal_string_20;
char *__gen_e_acsl_literal_string_17;
char *__gen_e_acsl_literal_string_18;
extern int __e_acsl_sound_verdict;

/*@ exits status: \exit_status ≡ \old(status);
    ensures never_terminates: \false;
    
    assigns \exit_status \from status;
 */
void __gen_e_acsl_exit(int status);

/*@ ensures
      result_null_or_valid_fd:
        \result ≡ \null ∨ \subset(\result, &__fc_fopen[0 .. 16 - 1]);
    assigns \result;
    assigns \result \from __fc_p_fopen;
 */
FILE *__gen_e_acsl_tmpfile(void);

/*@ requires valid_stream: \valid(stream);
    ensures result_zero_or_EOF: \result ≡ 0 ∨ \result ≡ -1;
    assigns \result;
    assigns \result \from (indirect: stream), (indirect: *stream);
 */
int __gen_e_acsl_fclose(FILE *stream);

/*@ ensures result_ok_or_error: \result ≡ -1 ∨ \result ≥ 0;
    ensures
      initialization: stat_loc_init_on_success:
        \result ≥ 0 ∧ \old(stat_loc) ≢ \null ⇒
        \initialized(\old(stat_loc));
    assigns \result, *stat_loc;
    assigns \result \from (indirect: options);
    assigns *stat_loc \from (indirect: options);
    
    behavior stat_loc_null:
      assumes stat_loc_null: stat_loc ≡ \null;
      assigns \result;
      assigns \result \from \nothing;
    
    behavior stat_loc_non_null:
      assumes stat_loc_non_null: stat_loc ≢ \null;
      requires valid_stat_loc: \valid(stat_loc);
 */
pid_t __gen_e_acsl_waitpid(pid_t pid, int *stat_loc, int options);

/*@ ensures
      result_ok_child_or_error:
        \result ≡ 0 ∨ \result > 0 ∨ \result ≡ -1;
    assigns \result;
    assigns \result \from \nothing;
 */
pid_t __gen_e_acsl_fork(void);

/*@ ensures
      result_ok_child_or_error:
        \result ≡ 0 ∨ \result > 0 ∨ \result ≡ -1;
    assigns \result;
    assigns \result \from \nothing;
 */
pid_t __gen_e_acsl_fork(void)
{
  pid_t __retres;
  __retres = fork();
  {
    int __gen_e_acsl_or;
    int __gen_e_acsl_or_2;
    if (__retres == 0) __gen_e_acsl_or = 1;
    else __gen_e_acsl_or = __retres > 0;
    if (__gen_e_acsl_or) __gen_e_acsl_or_2 = 1;
    else __gen_e_acsl_or_2 = __retres == -1;
    __e_acsl_assert(__gen_e_acsl_or_2,"Postcondition","fork",
                    "\\result == 0 || \\result > 0 || \\result == -1",
                    "FRAMAC_SHARE/libc/unistd.h",857);
    return __retres;
  }
}

/*@ ensures result_ok_or_error: \result ≡ -1 ∨ \result ≥ 0;
    ensures
      initialization: stat_loc_init_on_success:
        \result ≥ 0 ∧ \old(stat_loc) ≢ \null ⇒
        \initialized(\old(stat_loc));
    assigns \result, *stat_loc;
    assigns \result \from (indirect: options);
    assigns *stat_loc \from (indirect: options);
    
    behavior stat_loc_null:
      assumes stat_loc_null: stat_loc ≡ \null;
      assigns \result;
      assigns \result \from \nothing;
    
    behavior stat_loc_non_null:
      assumes stat_loc_non_null: stat_loc ≢ \null;
      requires valid_stat_loc: \valid(stat_loc);
 */
pid_t __gen_e_acsl_waitpid(pid_t pid, int *stat_loc, int options)
{
  int *__gen_e_acsl_at_2;
  int *__gen_e_acsl_at;
  __e_acsl_contract_t *__gen_e_acsl_contract;
  pid_t __retres;
  {
    int __gen_e_acsl_assumes_value;
    __e_acsl_store_block((void *)(& stat_loc),(size_t)8);
    __gen_e_acsl_contract = __e_acsl_contract_init((size_t)2);
    __e_acsl_contract_set_behavior_assumes(__gen_e_acsl_contract,(size_t)0,
                                           stat_loc == (int *)0);
    __e_acsl_contract_set_behavior_assumes(__gen_e_acsl_contract,(size_t)1,
                                           stat_loc != (int *)0);
    __gen_e_acsl_assumes_value = __e_acsl_contract_get_behavior_assumes
    ((__e_acsl_contract_t const *)__gen_e_acsl_contract,(size_t)1);
    if (__gen_e_acsl_assumes_value) {
      int __gen_e_acsl_valid;
      __gen_e_acsl_valid = __e_acsl_valid((void *)stat_loc,sizeof(int),
                                          (void *)stat_loc,
                                          (void *)(& stat_loc));
      __e_acsl_assert(__gen_e_acsl_valid,"Precondition","waitpid",
                      "stat_loc_non_null: \\valid(stat_loc)",
                      "FRAMAC_SHARE/libc/sys/wait.h",92);
    }
  }
  __gen_e_acsl_at_2 = stat_loc;
  __gen_e_acsl_at = stat_loc;
  __retres = waitpid(pid,stat_loc,options);
  {
    int __gen_e_acsl_or;
    int __gen_e_acsl_and;
    int __gen_e_acsl_implies;
    if (__retres == -1) __gen_e_acsl_or = 1;
    else __gen_e_acsl_or = __retres >= 0;
    __e_acsl_assert(__gen_e_acsl_or,"Postcondition","waitpid",
                    "\\result == -1 || \\result >= 0",
                    "FRAMAC_SHARE/libc/sys/wait.h",84);
    if (__retres >= 0) __gen_e_acsl_and = __gen_e_acsl_at != (int *)0;
    else __gen_e_acsl_and = 0;
    if (! __gen_e_acsl_and) __gen_e_acsl_implies = 1;
    else {
      int __gen_e_acsl_initialized;
      __gen_e_acsl_initialized = __e_acsl_initialized((void *)__gen_e_acsl_at_2,
                                                      sizeof(int));
      __gen_e_acsl_implies = __gen_e_acsl_initialized;
    }
    __e_acsl_assert(__gen_e_acsl_implies,"Postcondition","waitpid",
                    "\\result >= 0 && \\old(stat_loc) != \\null ==> \\initialized(\\old(stat_loc))",
                    "FRAMAC_SHARE/libc/sys/wait.h",86);
    __e_acsl_contract_clean(__gen_e_acsl_contract);
    __e_acsl_delete_block((void *)(& stat_loc));
    return __retres;
  }
}

/*@ requires valid_stream: \valid(stream);
    ensures result_zero_or_EOF: \result ≡ 0 ∨ \result ≡ -1;
    assigns \result;
    assigns \result \from (indirect: stream), (indirect: *stream);
 */
int __gen_e_acsl_fclose(FILE *stream)
{
  int __retres;
  {
    int __gen_e_acsl_valid;
    __e_acsl_store_block((void *)(& stream),(size_t)8);
    __gen_e_acsl_valid = __e_acsl_valid((void *)stream,sizeof(FILE),
                                        (void *)stream,(void *)(& stream));
    __e_acsl_assert(__gen_e_acsl_valid,"Precondition","fclose",
                    "\\valid(stream)","FRAMAC_SHARE/libc/stdio.h",120);
  }
  __retres = fclose(stream);
  {
    int __gen_e_acsl_or;
    if (__retres == 0) __gen_e_acsl_or = 1;
    else __gen_e_acsl_or = __retres == -1;
    __e_acsl_assert(__gen_e_acsl_or,"Postcondition","fclose",
                    "\\result == 0 || \\result == -1",
                    "FRAMAC_SHARE/libc/stdio.h",122);
    __e_acsl_delete_block((void *)(& stream));
    return __retres;
  }
}

/*@ ensures
      result_null_or_valid_fd:
        \result ≡ \null ∨ \subset(\result, &__fc_fopen[0 .. 16 - 1]);
    assigns \result;
    assigns \result \from __fc_p_fopen;
 */
FILE *__gen_e_acsl_tmpfile(void)
{
  FILE *__retres;
  __e_acsl_store_block((void *)(& __retres),(size_t)8);
  __retres = tmpfile();
  __e_acsl_delete_block((void *)(& __retres));
  return __retres;
}

/*@ exits status: \exit_status ≡ \old(status);
    ensures never_terminates: \false;
    
    assigns \exit_status \from status;
 */
void __gen_e_acsl_exit(int status)
{
  exit(status);
  __e_acsl_assert(0,"Postcondition","exit","\\false",
                  "FRAMAC_SHARE/libc/stdlib.h",473);
  return;
}

void __e_acsl_globals_init(void)
{
  static char __e_acsl_already_run = 0;
  if (! __e_acsl_already_run) {
    __e_acsl_already_run = 1;
    __gen_e_acsl_literal_string_31 = "tests/format/fprintf.c:46";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_31,
                         sizeof("tests/format/fprintf.c:46"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_31);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_31);
    __gen_e_acsl_literal_string_30 = "tests/format/fprintf.c:45";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_30,
                         sizeof("tests/format/fprintf.c:45"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_30);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_30);
    __gen_e_acsl_literal_string_29 = "tests/format/fprintf.c:44";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_29,
                         sizeof("tests/format/fprintf.c:44"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_29);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_29);
    __gen_e_acsl_literal_string_28 = "tests/format/fprintf.c:43";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_28,
                         sizeof("tests/format/fprintf.c:43"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_28);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_28);
    __gen_e_acsl_literal_string_27 = "tests/format/fprintf.c:42";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_27,
                         sizeof("tests/format/fprintf.c:42"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_27);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_27);
    __gen_e_acsl_literal_string_26 = "tests/format/fprintf.c:41";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_26,
                         sizeof("tests/format/fprintf.c:41"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_26);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_26);
    __gen_e_acsl_literal_string_25 = "tests/format/fprintf.c:38";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_25,
                         sizeof("tests/format/fprintf.c:38"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_25);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_25);
    __gen_e_acsl_literal_string_24 = "tests/format/fprintf.c:37";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_24,
                         sizeof("tests/format/fprintf.c:37"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_24);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_24);
    __gen_e_acsl_literal_string_23 = "tests/format/fprintf.c:36";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_23,
                         sizeof("tests/format/fprintf.c:36"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_23);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_23);
    __gen_e_acsl_literal_string_21 = "tests/format/fprintf.c:35";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_21,
                         sizeof("tests/format/fprintf.c:35"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_21);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_21);
    __gen_e_acsl_literal_string_19 = "tests/format/fprintf.c:34";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_19,
                         sizeof("tests/format/fprintf.c:34"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_19);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_19);
    __gen_e_acsl_literal_string_16 = "tests/format/fprintf.c:28";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_16,
                         sizeof("tests/format/fprintf.c:28"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_16);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_16);
    __gen_e_acsl_literal_string_15 = "tests/format/fprintf.c:27";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_15,
                         sizeof("tests/format/fprintf.c:27"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_15);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_15);
    __gen_e_acsl_literal_string_14 = "tests/format/fprintf.c:22";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_14,
                         sizeof("tests/format/fprintf.c:22"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_14);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_14);
    __gen_e_acsl_literal_string_13 = "tests/format/fprintf.c:21";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_13,
                         sizeof("tests/format/fprintf.c:21"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_13);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_13);
    __gen_e_acsl_literal_string_12 = "tests/format/fprintf.c:19";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_12,
                         sizeof("tests/format/fprintf.c:19"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_12);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_12);
    __gen_e_acsl_literal_string_9 = "tests/format/fprintf.c:16";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_9,
                         sizeof("tests/format/fprintf.c:16"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_9);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_9);
    __gen_e_acsl_literal_string_8 = "tests/format/fprintf.c:15";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_8,
                         sizeof("tests/format/fprintf.c:15"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_8);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_8);
    __gen_e_acsl_literal_string_11 = "foobar %s\n";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_11,
                         sizeof("foobar %s\n"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_11);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_11);
    __gen_e_acsl_literal_string_7 = "foobar\n";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_7,
                         sizeof("foobar\n"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_7);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_7);
    __gen_e_acsl_literal_string_10 = "foobar";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_10,
                         sizeof("foobar"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_10);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_10);
    __gen_e_acsl_literal_string = "TEST %d: ";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string,
                         sizeof("TEST %d: "));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string);
    __gen_e_acsl_literal_string_2 = "OK: Expected signal at %s\n";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_2,
                         sizeof("OK: Expected signal at %s\n"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_2);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_2);
    __gen_e_acsl_literal_string_3 = "OK: Expected execution at %s\n";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_3,
                         sizeof("OK: Expected execution at %s\n"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_3);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_3);
    __gen_e_acsl_literal_string_6 = "Hello world!";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_6,
                         sizeof("Hello world!"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_6);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_6);
    __gen_e_acsl_literal_string_5 = "FAIL: Unexpected signal at %s\n";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_5,
                         sizeof("FAIL: Unexpected signal at %s\n"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_5);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_5);
    __gen_e_acsl_literal_string_4 = "FAIL: Unexpected execution at %s\n";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_4,
                         sizeof("FAIL: Unexpected execution at %s\n"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_4);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_4);
    __gen_e_acsl_literal_string_22 = "123";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_22,
                         sizeof("123"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_22);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_22);
    __gen_e_acsl_literal_string_20 = "12";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_20,sizeof("12"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_20);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_20);
    __gen_e_acsl_literal_string_17 = "1";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_17,sizeof("1"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_17);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_17);
    __gen_e_acsl_literal_string_18 = "-%s-";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_18,
                         sizeof("-%s-"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_18);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_18);
  }
  return;
}

int main(int argc, char const **argv)
{
  int __retres;
  char template[256];
  char buf[5];
  __e_acsl_memory_init(& argc,(char ***)(& argv),(size_t)8);
  __e_acsl_globals_init();
  char *pstr = (char *)__gen_e_acsl_literal_string_6;
  {
    pid_t pid = __gen_e_acsl_fork();
    if (! pid) {
      __e_acsl_builtin_fprintf("",stdout,__gen_e_acsl_literal_string_7);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status;
      __e_acsl_store_block((void *)(& process_status),(size_t)4);
      __gen_e_acsl_waitpid(pid,& process_status,0);
      signal_eval(process_status,0,__gen_e_acsl_literal_string_8);
      __e_acsl_delete_block((void *)(& process_status));
    }
  }
  {
    pid_t pid_0 = __gen_e_acsl_fork();
    if (! pid_0) {
      __e_acsl_builtin_fprintf("",(FILE *)0,__gen_e_acsl_literal_string_7);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_0;
      __e_acsl_store_block((void *)(& process_status_0),(size_t)4);
      __gen_e_acsl_waitpid(pid_0,& process_status_0,0);
      signal_eval(process_status_0,1,__gen_e_acsl_literal_string_9);
      __e_acsl_delete_block((void *)(& process_status_0));
    }
  }
  FILE *fh = __gen_e_acsl_tmpfile();
  __e_acsl_store_block((void *)(& fh),(size_t)8);
  __e_acsl_full_init((void *)(& fh));
  if (fh) {
    {
      pid_t pid_1 = __gen_e_acsl_fork();
      if (! pid_1) {
        __e_acsl_builtin_fprintf("s",fh,__gen_e_acsl_literal_string_11,
                                 __gen_e_acsl_literal_string_10);
        __gen_e_acsl_exit(0);
      }
      else {
        int process_status_1;
        __e_acsl_store_block((void *)(& process_status_1),(size_t)4);
        __gen_e_acsl_waitpid(pid_1,& process_status_1,0);
        signal_eval(process_status_1,0,__gen_e_acsl_literal_string_12);
        __e_acsl_delete_block((void *)(& process_status_1));
      }
    }
    __gen_e_acsl_fclose(fh);
    {
      pid_t pid_2 = __gen_e_acsl_fork();
      if (! pid_2) {
        __e_acsl_builtin_fprintf("s",fh,__gen_e_acsl_literal_string_11,
                                 __gen_e_acsl_literal_string_10);
        __gen_e_acsl_exit(0);
      }
      else {
        int process_status_2;
        __e_acsl_store_block((void *)(& process_status_2),(size_t)4);
        __gen_e_acsl_waitpid(pid_2,& process_status_2,0);
        signal_eval(process_status_2,1,__gen_e_acsl_literal_string_13);
        __e_acsl_delete_block((void *)(& process_status_2));
      }
    }
    {
      pid_t pid_3 = __gen_e_acsl_fork();
      if (! pid_3) {
        __e_acsl_builtin_fprintf("s",(FILE *)(& argc),
                                 __gen_e_acsl_literal_string_11,
                                 __gen_e_acsl_literal_string_10);
        __gen_e_acsl_exit(0);
      }
      else {
        int process_status_3;
        __e_acsl_store_block((void *)(& process_status_3),(size_t)4);
        __gen_e_acsl_waitpid(pid_3,& process_status_3,0);
        signal_eval(process_status_3,1,__gen_e_acsl_literal_string_14);
        __e_acsl_delete_block((void *)(& process_status_3));
      }
    }
  }
  {
    pid_t pid_4 = __gen_e_acsl_fork();
    if (! pid_4) {
      __e_acsl_builtin_dprintf("",1,__gen_e_acsl_literal_string_7);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_4;
      __e_acsl_store_block((void *)(& process_status_4),(size_t)4);
      __gen_e_acsl_waitpid(pid_4,& process_status_4,0);
      signal_eval(process_status_4,0,__gen_e_acsl_literal_string_15);
      __e_acsl_delete_block((void *)(& process_status_4));
    }
  }
  {
    pid_t pid_5 = __gen_e_acsl_fork();
    if (! pid_5) {
      __e_acsl_builtin_dprintf("",3,__gen_e_acsl_literal_string_7);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_5;
      __e_acsl_store_block((void *)(& process_status_5),(size_t)4);
      __gen_e_acsl_waitpid(pid_5,& process_status_5,0);
      signal_eval(process_status_5,1,__gen_e_acsl_literal_string_16);
      __e_acsl_delete_block((void *)(& process_status_5));
    }
  }
  {
    pid_t pid_6 = __gen_e_acsl_fork();
    if (! pid_6) {
      __e_acsl_builtin_sprintf("s",buf,__gen_e_acsl_literal_string_18,
                               __gen_e_acsl_literal_string_17);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_6;
      __e_acsl_store_block((void *)(& process_status_6),(size_t)4);
      __gen_e_acsl_waitpid(pid_6,& process_status_6,0);
      signal_eval(process_status_6,0,__gen_e_acsl_literal_string_19);
      __e_acsl_delete_block((void *)(& process_status_6));
    }
  }
  {
    pid_t pid_7 = __gen_e_acsl_fork();
    if (! pid_7) {
      __e_acsl_builtin_sprintf("s",buf,__gen_e_acsl_literal_string_18,
                               __gen_e_acsl_literal_string_20);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_7;
      __e_acsl_store_block((void *)(& process_status_7),(size_t)4);
      __gen_e_acsl_waitpid(pid_7,& process_status_7,0);
      signal_eval(process_status_7,0,__gen_e_acsl_literal_string_21);
      __e_acsl_delete_block((void *)(& process_status_7));
    }
  }
  {
    pid_t pid_8 = __gen_e_acsl_fork();
    if (! pid_8) {
      __e_acsl_builtin_sprintf("s",buf,__gen_e_acsl_literal_string_18,
                               __gen_e_acsl_literal_string_22);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_8;
      __e_acsl_store_block((void *)(& process_status_8),(size_t)4);
      __gen_e_acsl_waitpid(pid_8,& process_status_8,0);
      signal_eval(process_status_8,1,__gen_e_acsl_literal_string_23);
      __e_acsl_delete_block((void *)(& process_status_8));
    }
  }
  {
    pid_t pid_9 = __gen_e_acsl_fork();
    if (! pid_9) {
      __e_acsl_builtin_sprintf("s",(char *)0,__gen_e_acsl_literal_string_18,
                               __gen_e_acsl_literal_string_22);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_9;
      __e_acsl_store_block((void *)(& process_status_9),(size_t)4);
      __gen_e_acsl_waitpid(pid_9,& process_status_9,0);
      signal_eval(process_status_9,1,__gen_e_acsl_literal_string_24);
      __e_acsl_delete_block((void *)(& process_status_9));
    }
  }
  {
    pid_t pid_10 = __gen_e_acsl_fork();
    if (! pid_10) {
      __e_acsl_builtin_sprintf("s",pstr,__gen_e_acsl_literal_string_18,
                               __gen_e_acsl_literal_string_22);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_10;
      __e_acsl_store_block((void *)(& process_status_10),(size_t)4);
      __gen_e_acsl_waitpid(pid_10,& process_status_10,0);
      signal_eval(process_status_10,1,__gen_e_acsl_literal_string_25);
      __e_acsl_delete_block((void *)(& process_status_10));
    }
  }
  {
    pid_t pid_11 = __gen_e_acsl_fork();
    if (! pid_11) {
      __e_acsl_builtin_snprintf("s",buf,(unsigned long)4,
                                __gen_e_acsl_literal_string_18,
                                __gen_e_acsl_literal_string_22);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_11;
      __e_acsl_store_block((void *)(& process_status_11),(size_t)4);
      __gen_e_acsl_waitpid(pid_11,& process_status_11,0);
      signal_eval(process_status_11,0,__gen_e_acsl_literal_string_26);
      __e_acsl_delete_block((void *)(& process_status_11));
    }
  }
  {
    pid_t pid_12 = __gen_e_acsl_fork();
    if (! pid_12) {
      __e_acsl_builtin_snprintf("s",buf,(unsigned long)5,
                                __gen_e_acsl_literal_string_18,
                                __gen_e_acsl_literal_string_22);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_12;
      __e_acsl_store_block((void *)(& process_status_12),(size_t)4);
      __gen_e_acsl_waitpid(pid_12,& process_status_12,0);
      signal_eval(process_status_12,0,__gen_e_acsl_literal_string_27);
      __e_acsl_delete_block((void *)(& process_status_12));
    }
  }
  {
    pid_t pid_13 = __gen_e_acsl_fork();
    if (! pid_13) {
      __e_acsl_builtin_snprintf("s",pstr,(unsigned long)6,
                                __gen_e_acsl_literal_string_18,
                                __gen_e_acsl_literal_string_22);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_13;
      __e_acsl_store_block((void *)(& process_status_13),(size_t)4);
      __gen_e_acsl_waitpid(pid_13,& process_status_13,0);
      signal_eval(process_status_13,1,__gen_e_acsl_literal_string_28);
      __e_acsl_delete_block((void *)(& process_status_13));
    }
  }
  {
    pid_t pid_14 = __gen_e_acsl_fork();
    if (! pid_14) {
      __e_acsl_builtin_snprintf("s",buf,(unsigned long)6,
                                __gen_e_acsl_literal_string_18,
                                __gen_e_acsl_literal_string_22);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_14;
      __e_acsl_store_block((void *)(& process_status_14),(size_t)4);
      __gen_e_acsl_waitpid(pid_14,& process_status_14,0);
      signal_eval(process_status_14,1,__gen_e_acsl_literal_string_29);
      __e_acsl_delete_block((void *)(& process_status_14));
    }
  }
  {
    pid_t pid_15 = __gen_e_acsl_fork();
    if (! pid_15) {
      __e_acsl_builtin_snprintf("s",(char *)0,(unsigned long)6,
                                __gen_e_acsl_literal_string_18,
                                __gen_e_acsl_literal_string_22);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_15;
      __e_acsl_store_block((void *)(& process_status_15),(size_t)4);
      __gen_e_acsl_waitpid(pid_15,& process_status_15,0);
      signal_eval(process_status_15,1,__gen_e_acsl_literal_string_30);
      __e_acsl_delete_block((void *)(& process_status_15));
    }
  }
  {
    pid_t pid_16 = __gen_e_acsl_fork();
    if (! pid_16) {
      __e_acsl_builtin_snprintf("s",(char *)0,(unsigned long)0,
                                __gen_e_acsl_literal_string_18,
                                __gen_e_acsl_literal_string_22);
      __gen_e_acsl_exit(0);
    }
    else {
      int process_status_16;
      __e_acsl_store_block((void *)(& process_status_16),(size_t)4);
      __gen_e_acsl_waitpid(pid_16,& process_status_16,0);
      signal_eval(process_status_16,0,__gen_e_acsl_literal_string_31);
      __e_acsl_delete_block((void *)(& process_status_16));
    }
  }
  __retres = 0;
  __e_acsl_delete_block((void *)(& fh));
  __e_acsl_memory_clean();
  return __retres;
}


