/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
int main(void)
{
  int __retres;
  int x = 0;
  if (x) {
    __e_acsl_assert(0,"Assertion","main","\\false",
                    "tests/constructs/false.i",6);
    /*@ assert \false; */ ;
  }
  __retres = 0;
  return __retres;
}


