typedef unsigned short u_int16_t;
typedef unsigned char u_char;

void *malloc(unsigned int size);

#define NS_GET32(l, cp) do { \
	u_char *t_cp = (u_char *)(cp); \
	(l) = ((u_int32_t)t_cp[0] << 24) \
	    | ((u_int32_t)t_cp[1] << 16) \
	    | ((u_int32_t)t_cp[2] << 8) \
	    | ((u_int32_t)t_cp[3]) \
	    ; \
	(cp) += 4; \
} while (0)

#define NS_GET16(s, cp) do { \
	u_char *t_cp = (u_char *)(cp); \
	(s) = ((u_int16_t)t_cp[0] << 8) \
	    | ((u_int16_t)t_cp[1]) \
	    ; \
	(cp) += 2; \
} while (0)


