/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
typedef unsigned char uint8;
int main(void)
{
  int __retres;
  uint8 x = (unsigned char)0;
  __e_acsl_assert((int)x == 0,"Assertion","main","x == 0",
                  "tests/constructs/typedef.i",9);
  /*@ assert x ≡ 0; */ ;
  __retres = 0;
  return __retres;
}


