#include "../constants.h"

int main ()
{
    struct sockaddr_un serv_adr;
    char               filename [FILENAME_SZ];

    /* server filename */
    filename[FILENAME_SZ-1] = EOS;
    
    /* initialize the server address structure */
    /* OK */
    r_strncpy (serv_adr.sun_path, filename, SUN_PATH_SZ-1);

    return 0;
}
