/* Generated by Frama-C */
#include "stddef.h"
#include "stdint.h"
#include "stdio.h"
#include "stdlib.h"
int *pfun(char c, int *p, int *p2, int *p3, int *p4, int i)
{
  __e_acsl_store_block((void *)(& p),(size_t)8);
  __e_acsl_temporal_pull_parameter((void *)(& p),1U,8UL);
  int *q = p;
  __e_acsl_temporal_store_nreferent((void *)(& q),(void *)(& p));
  __e_acsl_store_block((void *)(& q),(size_t)8);
  __e_acsl_full_init((void *)(& q));
  __e_acsl_temporal_save_return((void *)(& q));
  __e_acsl_delete_block((void *)(& p));
  __e_acsl_delete_block((void *)(& q));
  return q;
}

int main(int argc, char const **argv)
{
  int __retres;
  __e_acsl_memory_init(& argc,(char ***)(& argv),(size_t)8);
  int a = 1;
  __e_acsl_store_block((void *)(& a),(size_t)4);
  __e_acsl_full_init((void *)(& a));
  int *p = & a;
  __e_acsl_temporal_store_nblock((void *)(& p),(void *)(& a));
  __e_acsl_store_block((void *)(& p),(size_t)8);
  __e_acsl_full_init((void *)(& p));
  uintptr_t addr = (unsigned long)(& addr);
  __e_acsl_full_init((void *)(& p));
  __e_acsl_temporal_reset_parameters();
  __e_acsl_temporal_reset_return();
  __e_acsl_temporal_save_nreferent_parameter((void *)(& p),1U);
  __e_acsl_temporal_save_nblock_parameter((void *)(& a),2U);
  p = pfun((char)'a',p,& a,(int *)0,(int *)addr,2);
  __e_acsl_temporal_pull_return((void *)(& p));
  {
    int __gen_e_acsl_initialized;
    int __gen_e_acsl_and;
    __gen_e_acsl_initialized = __e_acsl_initialized((void *)(& p),
                                                    sizeof(int *));
    if (__gen_e_acsl_initialized) {
      int __gen_e_acsl_valid;
      __gen_e_acsl_valid = __e_acsl_valid((void *)p,sizeof(int),(void *)p,
                                          (void *)(& p));
      __gen_e_acsl_and = __gen_e_acsl_valid;
    }
    else __gen_e_acsl_and = 0;
    __e_acsl_assert(__gen_e_acsl_and,"Assertion","main","\\valid(p)",
                    "tests/temporal/t_fun_ptr.c",28);
  }
  /*@ assert \valid(p); */ ;
  __retres = 0;
  __e_acsl_delete_block((void *)(& p));
  __e_acsl_delete_block((void *)(& a));
  __e_acsl_memory_clean();
  return __retres;
}


