/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
int main(void)
{
  int __retres;
  __e_acsl_memory_init((int *)0,(char ***)0,(size_t)8);
  int i = 0;
  int arr[4] = {1, 2, 3, 4};
  __e_acsl_store_block((void *)(arr),(size_t)16);
  __e_acsl_full_init((void *)(& arr));
  int arr2[4] = {1, 2, 3, 4};
  if (! ((unsigned long)(& arr) + sizeof(arr) == (unsigned long)(& arr2))) {
    __retres = 0;
    goto return_label;
  }
  int *q = arr;
  __e_acsl_temporal_store_nblock((void *)(& q),(void *)(arr));
  __e_acsl_store_block((void *)(& q),(size_t)8);
  __e_acsl_full_init((void *)(& q));
  while (i < 4) {
    {
      int __gen_e_acsl_initialized;
      int __gen_e_acsl_and;
      __gen_e_acsl_initialized = __e_acsl_initialized((void *)(& q),
                                                      sizeof(int *));
      if (__gen_e_acsl_initialized) {
        int __gen_e_acsl_valid;
        __gen_e_acsl_valid = __e_acsl_valid((void *)q,sizeof(int),(void *)q,
                                            (void *)(& q));
        __gen_e_acsl_and = __gen_e_acsl_valid;
      }
      else __gen_e_acsl_and = 0;
      __e_acsl_assert(__gen_e_acsl_and,"Assertion","main","\\valid(q)",
                      "tests/temporal/t_while.c",28);
    }
    /*@ assert \valid(q); */ ;
    __e_acsl_initialize((void *)q,sizeof(int));
    *q = 1;
    i ++;
    __e_acsl_full_init((void *)(& q));
    __e_acsl_temporal_store_nreferent((void *)(& q),(void *)(& q));
    q ++;
  }
  {
    int __gen_e_acsl_initialized_2;
    int __gen_e_acsl_and_2;
    __gen_e_acsl_initialized_2 = __e_acsl_initialized((void *)(& q),
                                                      sizeof(int *));
    if (__gen_e_acsl_initialized_2) {
      int __gen_e_acsl_valid_2;
      __gen_e_acsl_valid_2 = __e_acsl_valid((void *)q,sizeof(int),(void *)q,
                                            (void *)(& q));
      __gen_e_acsl_and_2 = __gen_e_acsl_valid_2;
    }
    else __gen_e_acsl_and_2 = 0;
    __e_acsl_assert(! __gen_e_acsl_and_2,"Assertion","main","!\\valid(q)",
                    "tests/temporal/t_while.c",36);
  }
  /*@ assert ¬\valid(q); */ ;
  __retres = 0;
  return_label: __e_acsl_store_block_duplicate((void *)(& q),(size_t)8);
  __e_acsl_delete_block((void *)(& q));
  __e_acsl_delete_block((void *)(arr));
  __e_acsl_memory_clean();
  return __retres;
}


