(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2020                                               *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(** Computation of dominators. *)
open Cil_types;;

val get_idom: stmt -> stmt option
(** Immediate dominator of the statement. *)

val dominates: stmt -> stmt -> bool
(** [dominates a b] tells whether [a] dominates [b]. *)

val nearest_common_ancestor: stmt list -> stmt
(** Finds the statement lowest in the function that dominates
    all the statements in the list passed as argument. The list must not
    be empty, and must contain statements that are all in the same function. *)

(*
Local Variables:
compile-command: "make -C ../../.."
End:
*)
