/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
char const tab[];
char t[10];
void __e_acsl_globals_init(void)
{
  static char __e_acsl_already_run = 0;
  if (! __e_acsl_already_run) {
    __e_acsl_already_run = 1;
    __e_acsl_store_block((void *)(t),(size_t)10);
    __e_acsl_full_init((void *)(& t));
  }
  return;
}

void __e_acsl_globals_clean(void)
{
  __e_acsl_delete_block((void *)(t));
}

int main(void)
{
  int __retres;
  __e_acsl_memory_init((int *)0,(char ***)0,(size_t)8);
  __e_acsl_globals_init();
  char *p = (char *)(tab);
  __e_acsl_store_block((void *)(& p),(size_t)8);
  __e_acsl_full_init((void *)(& p));
  {
    int __gen_e_acsl_size;
    int __gen_e_acsl_if;
    int __gen_e_acsl_valid;
    __gen_e_acsl_size = 1 * ((9 - 0) + 1);
    if (__gen_e_acsl_size <= 0) __gen_e_acsl_if = 0;
    else __gen_e_acsl_if = __gen_e_acsl_size;
    __gen_e_acsl_valid = __e_acsl_valid((void *)(p + 1 * 0),
                                        (size_t)__gen_e_acsl_if,(void *)p,
                                        (void *)(& p));
    __e_acsl_assert(! __gen_e_acsl_valid,"Assertion","main",
                    "!\\valid(p + (0 .. 9))","tests/bts/bts2406.c",10);
  }
  /*@ assert ¬\valid(p + (0 .. 9)); */ ;
  {
    int __gen_e_acsl_size_2;
    int __gen_e_acsl_if_2;
    int __gen_e_acsl_valid_2;
    __gen_e_acsl_size_2 = 1 * ((9 - 0) + 1);
    if (__gen_e_acsl_size_2 <= 0) __gen_e_acsl_if_2 = 0;
    else __gen_e_acsl_if_2 = __gen_e_acsl_size_2;
    __gen_e_acsl_valid_2 = __e_acsl_valid((void *)(& t + 1 * 0),
                                          (size_t)__gen_e_acsl_if_2,
                                          (void *)(& t),(void *)0);
    __e_acsl_assert(__gen_e_acsl_valid_2,"Assertion","main",
                    "\\valid(&t[0 .. 9])","tests/bts/bts2406.c",11);
  }
  /*@ assert \valid(&t[0 .. 9]); */ ;
  __retres = 0;
  __e_acsl_delete_block((void *)(& p));
  __e_acsl_globals_clean();
  __e_acsl_memory_clean();
  return __retres;
}


