/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
int main(int argc, char const **argv)
{
  int __retres;
  __e_acsl_memory_init(& argc,(char ***)(& argv),(size_t)8);
  char a __attribute__((__aligned__(4))) = (char)'1';
  __e_acsl_store_block((void *)(& a),(size_t)1);
  __e_acsl_full_init((void *)(& a));
  char b __attribute__((__aligned__(4))) = (char)'2';
  __e_acsl_store_block((void *)(& b),(size_t)1);
  __e_acsl_full_init((void *)(& b));
  char *p = & a;
  __e_acsl_temporal_store_nblock((void *)(& p),(void *)(& a));
  __e_acsl_store_block((void *)(& p),(size_t)8);
  __e_acsl_full_init((void *)(& p));
  char *q = & b;
  __e_acsl_temporal_store_nblock((void *)(& q),(void *)(& b));
  __e_acsl_store_block((void *)(& q),(size_t)8);
  __e_acsl_full_init((void *)(& q));
  __e_acsl_full_init((void *)(& p));
  __e_acsl_temporal_store_nreferent((void *)(& p),(void *)(& q));
  p = q;
  {
    int __gen_e_acsl_initialized;
    int __gen_e_acsl_and;
    __gen_e_acsl_initialized = __e_acsl_initialized((void *)(& p),
                                                    sizeof(char *));
    if (__gen_e_acsl_initialized) {
      int __gen_e_acsl_valid;
      __gen_e_acsl_valid = __e_acsl_valid((void *)p,sizeof(char),(void *)p,
                                          (void *)(& p));
      __gen_e_acsl_and = __gen_e_acsl_valid;
    }
    else __gen_e_acsl_and = 0;
    __e_acsl_assert(__gen_e_acsl_and,"Assertion","main","\\valid(p)",
                    "tests/temporal/t_char.c",17);
  }
  /*@ assert \valid(p); */ ;
  {
    int __gen_e_acsl_initialized_2;
    int __gen_e_acsl_and_2;
    __gen_e_acsl_initialized_2 = __e_acsl_initialized((void *)(& q),
                                                      sizeof(char *));
    if (__gen_e_acsl_initialized_2) {
      int __gen_e_acsl_valid_2;
      __gen_e_acsl_valid_2 = __e_acsl_valid((void *)q,sizeof(char),(void *)q,
                                            (void *)(& q));
      __gen_e_acsl_and_2 = __gen_e_acsl_valid_2;
    }
    else __gen_e_acsl_and_2 = 0;
    __e_acsl_assert(__gen_e_acsl_and_2,"Assertion","main","\\valid(q)",
                    "tests/temporal/t_char.c",18);
  }
  /*@ assert \valid(q); */ ;
  __retres = 0;
  __e_acsl_delete_block((void *)(& q));
  __e_acsl_delete_block((void *)(& p));
  __e_acsl_delete_block((void *)(& b));
  __e_acsl_delete_block((void *)(& a));
  __e_acsl_memory_clean();
  return __retres;
}


