/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
#include "stdlib.h"
int A = 0;
int B;
void __e_acsl_globals_init(void)
{
  static char __e_acsl_already_run = 0;
  if (! __e_acsl_already_run) {
    __e_acsl_already_run = 1;
    __e_acsl_store_block((void *)(& B),(size_t)4);
    __e_acsl_full_init((void *)(& B));
    __e_acsl_store_block((void *)(& A),(size_t)4);
    __e_acsl_full_init((void *)(& A));
  }
  return;
}

void __e_acsl_globals_clean(void)
{
  __e_acsl_delete_block((void *)(& B));
  __e_acsl_delete_block((void *)(& A));
}

int main(void)
{
  int __retres;
  int b;
  long *r;
  long d[2];
  int dup[2];
  __e_acsl_memory_init((int *)0,(char ***)0,(size_t)8);
  __e_acsl_globals_init();
  __e_acsl_store_block((void *)(d),(size_t)16);
  __e_acsl_store_block((void *)(& r),(size_t)8);
  __e_acsl_store_block((void *)(& b),(size_t)4);
  int *p = & A;
  __e_acsl_store_block((void *)(& p),(size_t)8);
  __e_acsl_full_init((void *)(& p));
  int *q = & B;
  __e_acsl_store_block((void *)(& q),(size_t)8);
  __e_acsl_full_init((void *)(& q));
  __e_acsl_assert(1,"Assertion","main","\\initialized(&A)",
                  "tests/memory/initialized.c",16);
  /*@ assert \initialized(&A); */ ;
  __e_acsl_assert(1,"Assertion","main","\\initialized(&B)",
                  "tests/memory/initialized.c",17);
  /*@ assert \initialized(&B); */ ;
  {
    int __gen_e_acsl_initialized;
    __gen_e_acsl_initialized = __e_acsl_initialized((void *)p,sizeof(int));
    __e_acsl_assert(__gen_e_acsl_initialized,"Assertion","main",
                    "\\initialized(p)","tests/memory/initialized.c",18);
  }
  /*@ assert \initialized(p); */ ;
  {
    int __gen_e_acsl_initialized_2;
    __gen_e_acsl_initialized_2 = __e_acsl_initialized((void *)q,sizeof(int));
    __e_acsl_assert(__gen_e_acsl_initialized_2,"Assertion","main",
                    "\\initialized(q)","tests/memory/initialized.c",19);
  }
  /*@ assert \initialized(q); */ ;
  int a = 0;
  __e_acsl_store_block((void *)(& a),(size_t)4);
  __e_acsl_full_init((void *)(& a));
  long c[2] = {(long)1, (long)1};
  __e_acsl_store_block((void *)(c),(size_t)16);
  __e_acsl_full_init((void *)(& c));
  __e_acsl_full_init((void *)(& p));
  p = & a;
  __e_acsl_full_init((void *)(& q));
  q = & b;
  {
    int __gen_e_acsl_initialized_3;
    __gen_e_acsl_initialized_3 = __e_acsl_initialized((void *)(& a),
                                                      sizeof(int));
    __e_acsl_assert(__gen_e_acsl_initialized_3,"Assertion","main",
                    "\\initialized(&a)","tests/memory/initialized.c",30);
  }
  /*@ assert \initialized(&a); */ ;
  {
    int __gen_e_acsl_initialized_4;
    __gen_e_acsl_initialized_4 = __e_acsl_initialized((void *)(& b),
                                                      sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_4,"Assertion","main",
                    "!\\initialized(&b)","tests/memory/initialized.c",31);
  }
  /*@ assert ¬\initialized(&b); */ ;
  {
    int __gen_e_acsl_initialized_5;
    __gen_e_acsl_initialized_5 = __e_acsl_initialized((void *)p,sizeof(int));
    __e_acsl_assert(__gen_e_acsl_initialized_5,"Assertion","main",
                    "\\initialized(p)","tests/memory/initialized.c",32);
  }
  /*@ assert \initialized(p); */ ;
  {
    int __gen_e_acsl_initialized_6;
    __gen_e_acsl_initialized_6 = __e_acsl_initialized((void *)q,sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_6,"Assertion","main",
                    "!\\initialized(q)","tests/memory/initialized.c",33);
  }
  /*@ assert ¬\initialized(q); */ ;
  {
    int __gen_e_acsl_initialized_7;
    __gen_e_acsl_initialized_7 = __e_acsl_initialized((void *)(& c),
                                                      sizeof(long [2]));
    __e_acsl_assert(__gen_e_acsl_initialized_7,"Assertion","main",
                    "\\initialized(&c)","tests/memory/initialized.c",34);
  }
  /*@ assert \initialized(&c); */ ;
  {
    int __gen_e_acsl_initialized_8;
    __gen_e_acsl_initialized_8 = __e_acsl_initialized((void *)(& d),
                                                      sizeof(long [2]));
    __e_acsl_assert(! __gen_e_acsl_initialized_8,"Assertion","main",
                    "!\\initialized(&d)","tests/memory/initialized.c",35);
  }
  /*@ assert ¬\initialized(&d); */ ;
  __e_acsl_full_init((void *)(& b));
  b = 0;
  {
    int __gen_e_acsl_initialized_9;
    __gen_e_acsl_initialized_9 = __e_acsl_initialized((void *)q,sizeof(int));
    __e_acsl_assert(__gen_e_acsl_initialized_9,"Assertion","main",
                    "\\initialized(q)","tests/memory/initialized.c",39);
  }
  /*@ assert \initialized(q); */ ;
  {
    int __gen_e_acsl_initialized_10;
    __gen_e_acsl_initialized_10 = __e_acsl_initialized((void *)(& b),
                                                       sizeof(int));
    __e_acsl_assert(__gen_e_acsl_initialized_10,"Assertion","main",
                    "\\initialized(&b)","tests/memory/initialized.c",40);
  }
  /*@ assert \initialized(&b); */ ;
  __e_acsl_full_init((void *)(& r));
  r = d;
  {
    int __gen_e_acsl_initialized_11;
    __gen_e_acsl_initialized_11 = __e_acsl_initialized((void *)(d),
                                                       sizeof(long));
    __e_acsl_assert(! __gen_e_acsl_initialized_11,"Assertion","main",
                    "!\\initialized((long *)d)","tests/memory/initialized.c",
                    43);
  }
  /*@ assert ¬\initialized((long *)d); */ ;
  {
    int __gen_e_acsl_initialized_12;
    __gen_e_acsl_initialized_12 = __e_acsl_initialized((void *)(& d[1]),
                                                       sizeof(long));
    __e_acsl_assert(! __gen_e_acsl_initialized_12,"Assertion","main",
                    "!\\initialized(&d[1])","tests/memory/initialized.c",44);
  }
  /*@ assert ¬\initialized(&d[1]); */ ;
  {
    int __gen_e_acsl_initialized_13;
    __gen_e_acsl_initialized_13 = __e_acsl_initialized((void *)(& d),
                                                       sizeof(long [2]));
    __e_acsl_assert(! __gen_e_acsl_initialized_13,"Assertion","main",
                    "!\\initialized(&d)","tests/memory/initialized.c",45);
  }
  /*@ assert ¬\initialized(&d); */ ;
  {
    int __gen_e_acsl_initialized_14;
    __gen_e_acsl_initialized_14 = __e_acsl_initialized((void *)r,
                                                       sizeof(long));
    __e_acsl_assert(! __gen_e_acsl_initialized_14,"Assertion","main",
                    "!\\initialized(r)","tests/memory/initialized.c",46);
  }
  /*@ assert ¬\initialized(r); */ ;
  {
    int __gen_e_acsl_initialized_15;
    __gen_e_acsl_initialized_15 = __e_acsl_initialized((void *)(r + 1),
                                                       sizeof(long));
    __e_acsl_assert(! __gen_e_acsl_initialized_15,"Assertion","main",
                    "!\\initialized(r + 1)","tests/memory/initialized.c",47);
  }
  /*@ assert ¬\initialized(r + 1); */ ;
  __e_acsl_initialize((void *)(d),sizeof(long));
  d[0] = (long)1;
  {
    int __gen_e_acsl_initialized_16;
    __gen_e_acsl_initialized_16 = __e_acsl_initialized((void *)(d),
                                                       sizeof(long));
    __e_acsl_assert(__gen_e_acsl_initialized_16,"Assertion","main",
                    "\\initialized((long *)d)","tests/memory/initialized.c",
                    50);
  }
  /*@ assert \initialized((long *)d); */ ;
  {
    int __gen_e_acsl_initialized_17;
    __gen_e_acsl_initialized_17 = __e_acsl_initialized((void *)(& d[1]),
                                                       sizeof(long));
    __e_acsl_assert(! __gen_e_acsl_initialized_17,"Assertion","main",
                    "!\\initialized(&d[1])","tests/memory/initialized.c",51);
  }
  /*@ assert ¬\initialized(&d[1]); */ ;
  {
    int __gen_e_acsl_initialized_18;
    __gen_e_acsl_initialized_18 = __e_acsl_initialized((void *)(& d),
                                                       sizeof(long [2]));
    __e_acsl_assert(! __gen_e_acsl_initialized_18,"Assertion","main",
                    "!\\initialized(&d)","tests/memory/initialized.c",52);
  }
  /*@ assert ¬\initialized(&d); */ ;
  {
    int __gen_e_acsl_initialized_19;
    __gen_e_acsl_initialized_19 = __e_acsl_initialized((void *)r,
                                                       sizeof(long));
    __e_acsl_assert(__gen_e_acsl_initialized_19,"Assertion","main",
                    "\\initialized(r)","tests/memory/initialized.c",53);
  }
  /*@ assert \initialized(r); */ ;
  {
    int __gen_e_acsl_initialized_20;
    __gen_e_acsl_initialized_20 = __e_acsl_initialized((void *)(r + 1),
                                                       sizeof(long));
    __e_acsl_assert(! __gen_e_acsl_initialized_20,"Assertion","main",
                    "!\\initialized(r + 1)","tests/memory/initialized.c",54);
  }
  /*@ assert ¬\initialized(r + 1); */ ;
  __e_acsl_initialize((void *)(& d[1]),sizeof(long));
  d[1] = (long)1;
  {
    int __gen_e_acsl_initialized_21;
    __gen_e_acsl_initialized_21 = __e_acsl_initialized((void *)(d),
                                                       sizeof(long));
    __e_acsl_assert(__gen_e_acsl_initialized_21,"Assertion","main",
                    "\\initialized((long *)d)","tests/memory/initialized.c",
                    57);
  }
  /*@ assert \initialized((long *)d); */ ;
  {
    int __gen_e_acsl_initialized_22;
    __gen_e_acsl_initialized_22 = __e_acsl_initialized((void *)(& d[1]),
                                                       sizeof(long));
    __e_acsl_assert(__gen_e_acsl_initialized_22,"Assertion","main",
                    "\\initialized(&d[1])","tests/memory/initialized.c",58);
  }
  /*@ assert \initialized(&d[1]); */ ;
  {
    int __gen_e_acsl_initialized_23;
    __gen_e_acsl_initialized_23 = __e_acsl_initialized((void *)(& d),
                                                       sizeof(long [2]));
    __e_acsl_assert(__gen_e_acsl_initialized_23,"Assertion","main",
                    "\\initialized(&d)","tests/memory/initialized.c",59);
  }
  /*@ assert \initialized(&d); */ ;
  {
    int __gen_e_acsl_initialized_24;
    __gen_e_acsl_initialized_24 = __e_acsl_initialized((void *)r,
                                                       sizeof(long));
    __e_acsl_assert(__gen_e_acsl_initialized_24,"Assertion","main",
                    "\\initialized(r)","tests/memory/initialized.c",60);
  }
  /*@ assert \initialized(r); */ ;
  {
    int __gen_e_acsl_initialized_25;
    __gen_e_acsl_initialized_25 = __e_acsl_initialized((void *)(r + 1),
                                                       sizeof(long));
    __e_acsl_assert(__gen_e_acsl_initialized_25,"Assertion","main",
                    "\\initialized(r + 1)","tests/memory/initialized.c",61);
  }
  /*@ assert \initialized(r + 1); */ ;
  __e_acsl_full_init((void *)(& p));
  p = (int *)malloc(sizeof(int *));
  {
    int __gen_e_acsl_initialized_26;
    __gen_e_acsl_initialized_26 = __e_acsl_initialized((void *)p,sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_26,"Assertion","main",
                    "!\\initialized(p)","tests/memory/initialized.c",65);
  }
  /*@ assert ¬\initialized(p); */ ;
  __e_acsl_full_init((void *)(& q));
  q = (int *)calloc((unsigned long)1,sizeof(int));
  {
    int __gen_e_acsl_initialized_27;
    __gen_e_acsl_initialized_27 = __e_acsl_initialized((void *)q,sizeof(int));
    __e_acsl_assert(__gen_e_acsl_initialized_27,"Assertion","main",
                    "\\initialized(q)","tests/memory/initialized.c",69);
  }
  /*@ assert \initialized(q); */ ;
  __e_acsl_full_init((void *)(& q));
  q = (int *)realloc((void *)q,(unsigned long)2 * sizeof(int));
  {
    int __gen_e_acsl_initialized_28;
    __gen_e_acsl_initialized_28 = __e_acsl_initialized((void *)q,sizeof(int));
    __e_acsl_assert(__gen_e_acsl_initialized_28,"Assertion","main",
                    "\\initialized(q)","tests/memory/initialized.c",74);
  }
  /*@ assert \initialized(q); */ ;
  __e_acsl_full_init((void *)(& q));
  q ++;
  {
    int __gen_e_acsl_initialized_29;
    __gen_e_acsl_initialized_29 = __e_acsl_initialized((void *)q,sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_29,"Assertion","main",
                    "!\\initialized(q)","tests/memory/initialized.c",76);
  }
  /*@ assert ¬\initialized(q); */ ;
  __e_acsl_full_init((void *)(& q));
  q --;
  free((void *)p);
  free((void *)q);
  {
    int __gen_e_acsl_initialized_30;
    /*@ assert Eva: dangling_pointer: ¬\dangling(&p); */
    __gen_e_acsl_initialized_30 = __e_acsl_initialized((void *)p,sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_30,"Assertion","main",
                    "!\\initialized(p)","tests/memory/initialized.c",84);
  }
  /*@ assert ¬\initialized(p); */ ;
  {
    int __gen_e_acsl_initialized_31;
    __gen_e_acsl_initialized_31 = __e_acsl_initialized((void *)q,sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_31,"Assertion","main",
                    "!\\initialized(q)","tests/memory/initialized.c",85);
  }
  /*@ assert ¬\initialized(q); */ ;
  __e_acsl_full_init((void *)(& q));
  q = (int *)calloc((unsigned long)3,sizeof(int));
  {
    int __gen_e_acsl_size;
    int __gen_e_acsl_if;
    int __gen_e_acsl_initialized_32;
    __gen_e_acsl_size = 4 * ((2 - 0) + 1);
    if (__gen_e_acsl_size <= 0) __gen_e_acsl_if = 0;
    else __gen_e_acsl_if = __gen_e_acsl_size;
    __gen_e_acsl_initialized_32 = __e_acsl_initialized((void *)((char *)q + 
                                                                4 * 0),
                                                       (size_t)__gen_e_acsl_if);
    __e_acsl_assert(__gen_e_acsl_initialized_32,"Assertion","main",
                    "\\initialized(q + (0 .. 2))",
                    "tests/memory/initialized.c",90);
  }
  /*@ assert \initialized(q + (0 .. 2)); */ ;
  __e_acsl_full_init((void *)(& q));
  q = (int *)realloc((void *)q,(unsigned long)6 * sizeof(int));
  {
    int __gen_e_acsl_size_2;
    int __gen_e_acsl_if_2;
    int __gen_e_acsl_initialized_33;
    __gen_e_acsl_size_2 = 4 * ((2 - 0) + 1);
    if (__gen_e_acsl_size_2 <= 0) __gen_e_acsl_if_2 = 0;
    else __gen_e_acsl_if_2 = __gen_e_acsl_size_2;
    __gen_e_acsl_initialized_33 = __e_acsl_initialized((void *)((char *)q + 
                                                                4 * 0),
                                                       (size_t)__gen_e_acsl_if_2);
    __e_acsl_assert(__gen_e_acsl_initialized_33,"Assertion","main",
                    "\\initialized(q + (0 .. 2))",
                    "tests/memory/initialized.c",92);
  }
  /*@ assert \initialized(q + (0 .. 2)); */ ;
  {
    int __gen_e_acsl_initialized_34;
    __gen_e_acsl_initialized_34 = __e_acsl_initialized((void *)(q + 3),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_34,"Assertion","main",
                    "!\\initialized(q + 3)","tests/memory/initialized.c",93);
  }
  /*@ assert ¬\initialized(q + 3); */ ;
  {
    int __gen_e_acsl_initialized_35;
    __gen_e_acsl_initialized_35 = __e_acsl_initialized((void *)(q + 4),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_35,"Assertion","main",
                    "!\\initialized(q + 4)","tests/memory/initialized.c",94);
  }
  /*@ assert ¬\initialized(q + 4); */ ;
  {
    int __gen_e_acsl_initialized_36;
    __gen_e_acsl_initialized_36 = __e_acsl_initialized((void *)(q + 5),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_36,"Assertion","main",
                    "!\\initialized(q + 5)","tests/memory/initialized.c",95);
  }
  /*@ assert ¬\initialized(q + 5); */ ;
  free((void *)q);
  __e_acsl_full_init((void *)(& q));
  q = (int *)calloc((unsigned long)7,sizeof(int));
  {
    int __gen_e_acsl_size_3;
    int __gen_e_acsl_if_3;
    int __gen_e_acsl_initialized_37;
    __gen_e_acsl_size_3 = 4 * ((6 - 0) + 1);
    if (__gen_e_acsl_size_3 <= 0) __gen_e_acsl_if_3 = 0;
    else __gen_e_acsl_if_3 = __gen_e_acsl_size_3;
    __gen_e_acsl_initialized_37 = __e_acsl_initialized((void *)((char *)q + 
                                                                4 * 0),
                                                       (size_t)__gen_e_acsl_if_3);
    __e_acsl_assert(__gen_e_acsl_initialized_37,"Assertion","main",
                    "\\initialized(q + (0 .. 6))",
                    "tests/memory/initialized.c",98);
  }
  /*@ assert \initialized(q + (0 .. 6)); */ ;
  __e_acsl_full_init((void *)(& q));
  q = (int *)realloc((void *)q,(unsigned long)8 * sizeof(int));
  {
    int __gen_e_acsl_size_4;
    int __gen_e_acsl_if_4;
    int __gen_e_acsl_initialized_38;
    __gen_e_acsl_size_4 = 4 * ((6 - 0) + 1);
    if (__gen_e_acsl_size_4 <= 0) __gen_e_acsl_if_4 = 0;
    else __gen_e_acsl_if_4 = __gen_e_acsl_size_4;
    __gen_e_acsl_initialized_38 = __e_acsl_initialized((void *)((char *)q + 
                                                                4 * 0),
                                                       (size_t)__gen_e_acsl_if_4);
    __e_acsl_assert(__gen_e_acsl_initialized_38,"Assertion","main",
                    "\\initialized(q + (0 .. 6))",
                    "tests/memory/initialized.c",100);
  }
  /*@ assert \initialized(q + (0 .. 6)); */ ;
  {
    int __gen_e_acsl_initialized_39;
    __gen_e_acsl_initialized_39 = __e_acsl_initialized((void *)(q + 7),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_39,"Assertion","main",
                    "!\\initialized(q + 7)","tests/memory/initialized.c",101);
  }
  /*@ assert ¬\initialized(q + 7); */ ;
  __e_acsl_full_init((void *)(& q));
  q = (int *)realloc((void *)q,(unsigned long)10 * sizeof(int));
  {
    int __gen_e_acsl_size_5;
    int __gen_e_acsl_if_5;
    int __gen_e_acsl_initialized_40;
    __gen_e_acsl_size_5 = 4 * ((6 - 0) + 1);
    if (__gen_e_acsl_size_5 <= 0) __gen_e_acsl_if_5 = 0;
    else __gen_e_acsl_if_5 = __gen_e_acsl_size_5;
    __gen_e_acsl_initialized_40 = __e_acsl_initialized((void *)((char *)q + 
                                                                4 * 0),
                                                       (size_t)__gen_e_acsl_if_5);
    __e_acsl_assert(__gen_e_acsl_initialized_40,"Assertion","main",
                    "\\initialized(q + (0 .. 6))",
                    "tests/memory/initialized.c",103);
  }
  /*@ assert \initialized(q + (0 .. 6)); */ ;
  {
    int __gen_e_acsl_initialized_41;
    __gen_e_acsl_initialized_41 = __e_acsl_initialized((void *)(q + 7),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_41,"Assertion","main",
                    "!\\initialized(q + 7)","tests/memory/initialized.c",104);
  }
  /*@ assert ¬\initialized(q + 7); */ ;
  {
    int __gen_e_acsl_initialized_42;
    __gen_e_acsl_initialized_42 = __e_acsl_initialized((void *)(q + 8),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_42,"Assertion","main",
                    "!\\initialized(q + 8)","tests/memory/initialized.c",105);
  }
  /*@ assert ¬\initialized(q + 8); */ ;
  {
    int __gen_e_acsl_initialized_43;
    __gen_e_acsl_initialized_43 = __e_acsl_initialized((void *)(q + 9),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_43,"Assertion","main",
                    "!\\initialized(q + 9)","tests/memory/initialized.c",106);
  }
  /*@ assert ¬\initialized(q + 9); */ ;
  free((void *)q);
  __e_acsl_full_init((void *)(& q));
  q = (int *)calloc((unsigned long)2,sizeof(int));
  {
    int __gen_e_acsl_size_6;
    int __gen_e_acsl_if_6;
    int __gen_e_acsl_initialized_44;
    __gen_e_acsl_size_6 = 4 * ((1 - 0) + 1);
    if (__gen_e_acsl_size_6 <= 0) __gen_e_acsl_if_6 = 0;
    else __gen_e_acsl_if_6 = __gen_e_acsl_size_6;
    __gen_e_acsl_initialized_44 = __e_acsl_initialized((void *)((char *)q + 
                                                                4 * 0),
                                                       (size_t)__gen_e_acsl_if_6);
    __e_acsl_assert(__gen_e_acsl_initialized_44,"Assertion","main",
                    "\\initialized(q + (0 .. 1))",
                    "tests/memory/initialized.c",109);
  }
  /*@ assert \initialized(q + (0 .. 1)); */ ;
  __e_acsl_full_init((void *)(& q));
  q = (int *)realloc((void *)q,(unsigned long)4 * sizeof(int));
  {
    int __gen_e_acsl_size_7;
    int __gen_e_acsl_if_7;
    int __gen_e_acsl_initialized_45;
    __gen_e_acsl_size_7 = 4 * ((1 - 0) + 1);
    if (__gen_e_acsl_size_7 <= 0) __gen_e_acsl_if_7 = 0;
    else __gen_e_acsl_if_7 = __gen_e_acsl_size_7;
    __gen_e_acsl_initialized_45 = __e_acsl_initialized((void *)((char *)q + 
                                                                4 * 0),
                                                       (size_t)__gen_e_acsl_if_7);
    __e_acsl_assert(__gen_e_acsl_initialized_45,"Assertion","main",
                    "\\initialized(q + (0 .. 1))",
                    "tests/memory/initialized.c",111);
  }
  /*@ assert \initialized(q + (0 .. 1)); */ ;
  {
    int __gen_e_acsl_initialized_46;
    __gen_e_acsl_initialized_46 = __e_acsl_initialized((void *)(q + 2),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_46,"Assertion","main",
                    "!\\initialized(q + 2)","tests/memory/initialized.c",112);
  }
  /*@ assert ¬\initialized(q + 2); */ ;
  {
    int __gen_e_acsl_initialized_47;
    __gen_e_acsl_initialized_47 = __e_acsl_initialized((void *)(q + 3),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_47,"Assertion","main",
                    "!\\initialized(q + 3)","tests/memory/initialized.c",113);
  }
  /*@ assert ¬\initialized(q + 3); */ ;
  free((void *)q);
  __e_acsl_full_init((void *)(& q));
  q = (int *)calloc((unsigned long)6,sizeof(int));
  {
    int __gen_e_acsl_size_8;
    int __gen_e_acsl_if_8;
    int __gen_e_acsl_initialized_48;
    __gen_e_acsl_size_8 = 4 * ((5 - 0) + 1);
    if (__gen_e_acsl_size_8 <= 0) __gen_e_acsl_if_8 = 0;
    else __gen_e_acsl_if_8 = __gen_e_acsl_size_8;
    __gen_e_acsl_initialized_48 = __e_acsl_initialized((void *)((char *)q + 
                                                                4 * 0),
                                                       (size_t)__gen_e_acsl_if_8);
    __e_acsl_assert(__gen_e_acsl_initialized_48,"Assertion","main",
                    "\\initialized(q + (0 .. 5))",
                    "tests/memory/initialized.c",116);
  }
  /*@ assert \initialized(q + (0 .. 5)); */ ;
  __e_acsl_full_init((void *)(& q));
  q = (int *)realloc((void *)q,(unsigned long)3 * sizeof(int));
  {
    int __gen_e_acsl_size_9;
    int __gen_e_acsl_if_9;
    int __gen_e_acsl_initialized_49;
    __gen_e_acsl_size_9 = 4 * ((2 - 0) + 1);
    if (__gen_e_acsl_size_9 <= 0) __gen_e_acsl_if_9 = 0;
    else __gen_e_acsl_if_9 = __gen_e_acsl_size_9;
    __gen_e_acsl_initialized_49 = __e_acsl_initialized((void *)((char *)q + 
                                                                4 * 0),
                                                       (size_t)__gen_e_acsl_if_9);
    __e_acsl_assert(__gen_e_acsl_initialized_49,"Assertion","main",
                    "\\initialized(q + (0 .. 2))",
                    "tests/memory/initialized.c",118);
  }
  /*@ assert \initialized(q + (0 .. 2)); */ ;
  free((void *)q);
  __e_acsl_full_init((void *)(& q));
  q = (int *)malloc((unsigned long)6 * sizeof(int));
  {
    int __gen_e_acsl_initialized_50;
    __gen_e_acsl_initialized_50 = __e_acsl_initialized((void *)(q + 0),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_50,"Assertion","main",
                    "!\\initialized(q + 0)","tests/memory/initialized.c",121);
  }
  /*@ assert ¬\initialized(q + 0); */ ;
  {
    int __gen_e_acsl_initialized_51;
    __gen_e_acsl_initialized_51 = __e_acsl_initialized((void *)(q + 1),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_51,"Assertion","main",
                    "!\\initialized(q + 1)","tests/memory/initialized.c",122);
  }
  /*@ assert ¬\initialized(q + 1); */ ;
  {
    int __gen_e_acsl_initialized_52;
    __gen_e_acsl_initialized_52 = __e_acsl_initialized((void *)(q + 2),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_52,"Assertion","main",
                    "!\\initialized(q + 2)","tests/memory/initialized.c",123);
  }
  /*@ assert ¬\initialized(q + 2); */ ;
  {
    int __gen_e_acsl_initialized_53;
    __gen_e_acsl_initialized_53 = __e_acsl_initialized((void *)(q + 3),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_53,"Assertion","main",
                    "!\\initialized(q + 3)","tests/memory/initialized.c",124);
  }
  /*@ assert ¬\initialized(q + 3); */ ;
  {
    int __gen_e_acsl_initialized_54;
    __gen_e_acsl_initialized_54 = __e_acsl_initialized((void *)(q + 4),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_54,"Assertion","main",
                    "!\\initialized(q + 4)","tests/memory/initialized.c",125);
  }
  /*@ assert ¬\initialized(q + 4); */ ;
  {
    int __gen_e_acsl_initialized_55;
    __gen_e_acsl_initialized_55 = __e_acsl_initialized((void *)(q + 5),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_55,"Assertion","main",
                    "!\\initialized(q + 5)","tests/memory/initialized.c",126);
  }
  /*@ assert ¬\initialized(q + 5); */ ;
  __e_acsl_full_init((void *)(& q));
  q = (int *)realloc((void *)q,(unsigned long)3 * sizeof(int));
  {
    int __gen_e_acsl_initialized_56;
    __gen_e_acsl_initialized_56 = __e_acsl_initialized((void *)(q + 0),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_56,"Assertion","main",
                    "!\\initialized(q + 0)","tests/memory/initialized.c",128);
  }
  /*@ assert ¬\initialized(q + 0); */ ;
  {
    int __gen_e_acsl_initialized_57;
    __gen_e_acsl_initialized_57 = __e_acsl_initialized((void *)(q + 1),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_57,"Assertion","main",
                    "!\\initialized(q + 1)","tests/memory/initialized.c",129);
  }
  /*@ assert ¬\initialized(q + 1); */ ;
  {
    int __gen_e_acsl_initialized_58;
    __gen_e_acsl_initialized_58 = __e_acsl_initialized((void *)(q + 2),
                                                       sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_58,"Assertion","main",
                    "!\\initialized(q + 2)","tests/memory/initialized.c",130);
  }
  /*@ assert ¬\initialized(q + 2); */ ;
  free((void *)q);
  __e_acsl_full_init((void *)(& q));
  q = (int *)(& q - 1024 * 5);
  __e_acsl_full_init((void *)(& q));
  q = (int *)128;
  {
    int __gen_e_acsl_initialized_59;
    __gen_e_acsl_initialized_59 = __e_acsl_initialized((void *)q,sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_59,"Assertion","main",
                    "!\\initialized(q)","tests/memory/initialized.c",139);
  }
  /*@ assert ¬\initialized(q); */ ;
  __e_acsl_full_init((void *)(& p));
  p = (int *)0;
  {
    int __gen_e_acsl_initialized_60;
    __gen_e_acsl_initialized_60 = __e_acsl_initialized((void *)p,sizeof(int));
    __e_acsl_assert(! __gen_e_acsl_initialized_60,"Assertion","main",
                    "!\\initialized(p)","tests/memory/initialized.c",142);
  }
  /*@ assert ¬\initialized(p); */ ;
  int size = 100;
  char *partsc = malloc((unsigned long)size * sizeof(char));
  char *partsi = malloc((unsigned long)size * sizeof(int));
  {
    int i = 0;
    while (i < size) {
      if (i % 2 != 0) *(partsc + i) = (char)'0';
      else *(partsi + i) = (char)0;
      i ++;
    }
  }
  {
    int i_0 = 0;
    while (i_0 < size) {
      if (i_0 % 2 != 0) ;
      i_0 ++;
    }
  }
  dup[0] = 1;
  dup[0] = 1;
  __retres = 0;
  __e_acsl_delete_block((void *)(d));
  __e_acsl_delete_block((void *)(c));
  __e_acsl_delete_block((void *)(& r));
  __e_acsl_delete_block((void *)(& b));
  __e_acsl_delete_block((void *)(& a));
  __e_acsl_delete_block((void *)(& q));
  __e_acsl_delete_block((void *)(& p));
  __e_acsl_globals_clean();
  __e_acsl_memory_clean();
  return __retres;
}


