/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
#include "stdlib.h"
char *__gen_e_acsl_literal_string;
void __e_acsl_globals_init(void)
{
  static char __e_acsl_already_run = 0;
  if (! __e_acsl_already_run) {
    __e_acsl_already_run = 1;
    __gen_e_acsl_literal_string = ".";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string,sizeof("."));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string);
  }
  return;
}

int main(void)
{
  int __retres;
  __e_acsl_memory_init((int *)0,(char ***)0,(size_t)8);
  __e_acsl_globals_init();
  char *c = (char *)__gen_e_acsl_literal_string;
  __e_acsl_temporal_reset_parameters();
  __e_acsl_temporal_reset_return();
  char *p = malloc((unsigned long)256);
  __e_acsl_temporal_store_nblock((void *)(& p),(void *)*(& p));
  __e_acsl_store_block((void *)(& p),(size_t)8);
  __e_acsl_full_init((void *)(& p));
  char *q = (char *)0;
  __e_acsl_temporal_store_nblock((void *)(& q),(void *)0);
  __e_acsl_store_block((void *)(& q),(size_t)8);
  __e_acsl_full_init((void *)(& q));
  __e_acsl_full_init((void *)(& q));
  __e_acsl_temporal_reset_parameters();
  __e_acsl_temporal_reset_return();
  q = (char *)malloc((unsigned long)256);
  __e_acsl_temporal_store_nblock((void *)(& q),(void *)*(& q));
  {
    int __gen_e_acsl_initialized;
    int __gen_e_acsl_and;
    int __gen_e_acsl_and_3;
    __gen_e_acsl_initialized = __e_acsl_initialized((void *)(& q),
                                                    sizeof(char *));
    if (__gen_e_acsl_initialized) {
      int __gen_e_acsl_valid;
      __gen_e_acsl_valid = __e_acsl_valid((void *)q,sizeof(char),(void *)q,
                                          (void *)(& q));
      __gen_e_acsl_and = __gen_e_acsl_valid;
    }
    else __gen_e_acsl_and = 0;
    if (__gen_e_acsl_and) {
      int __gen_e_acsl_initialized_2;
      int __gen_e_acsl_and_2;
      __gen_e_acsl_initialized_2 = __e_acsl_initialized((void *)(& p),
                                                        sizeof(char *));
      if (__gen_e_acsl_initialized_2) {
        int __gen_e_acsl_valid_2;
        __gen_e_acsl_valid_2 = __e_acsl_valid((void *)p,sizeof(char),
                                              (void *)p,(void *)(& p));
        __gen_e_acsl_and_2 = __gen_e_acsl_valid_2;
      }
      else __gen_e_acsl_and_2 = 0;
      __gen_e_acsl_and_3 = __gen_e_acsl_and_2;
    }
    else __gen_e_acsl_and_3 = 0;
    __e_acsl_assert(__gen_e_acsl_and_3,"Assertion","main",
                    "\\valid(q) && \\valid(p)","tests/temporal/t_fun_lib.c",
                    16);
  }
  /*@ assert \valid(q) ∧ \valid(p); */ ;
  __e_acsl_temporal_reset_parameters();
  __e_acsl_temporal_reset_return();
  __e_acsl_temporal_save_nreferent_parameter((void *)(& q),1U);
  char *path = realpath((char const *)c,q);
  __e_acsl_temporal_store_nblock((void *)(& path),(void *)*(& path));
  __e_acsl_store_block((void *)(& path),(size_t)8);
  __e_acsl_full_init((void *)(& path));
  __e_acsl_full_init((void *)(& path));
  __e_acsl_temporal_reset_parameters();
  __e_acsl_temporal_reset_return();
  __e_acsl_temporal_save_nreferent_parameter((void *)(& q),1U);
  path = realpath((char const *)c,q);
  __e_acsl_temporal_store_nblock((void *)(& path),(void *)*(& path));
  {
    int __gen_e_acsl_initialized_3;
    int __gen_e_acsl_and_4;
    __gen_e_acsl_initialized_3 = __e_acsl_initialized((void *)(& path),
                                                      sizeof(char *));
    if (__gen_e_acsl_initialized_3) {
      int __gen_e_acsl_valid_3;
      __gen_e_acsl_valid_3 = __e_acsl_valid((void *)path,sizeof(char),
                                            (void *)path,(void *)(& path));
      __gen_e_acsl_and_4 = __gen_e_acsl_valid_3;
    }
    else __gen_e_acsl_and_4 = 0;
    __e_acsl_assert(__gen_e_acsl_and_4,"Assertion","main","\\valid(path)",
                    "tests/temporal/t_fun_lib.c",23);
  }
  /*@ assert \valid(path); */ ;
  __e_acsl_temporal_reset_parameters();
  __e_acsl_temporal_reset_return();
  __e_acsl_temporal_save_nreferent_parameter((void *)(& p),0U);
  free((void *)p);
  __e_acsl_temporal_reset_parameters();
  __e_acsl_temporal_reset_return();
  __e_acsl_temporal_save_nreferent_parameter((void *)(& path),0U);
  free((void *)path);
  {
    int __gen_e_acsl_initialized_4;
    int __gen_e_acsl_and_5;
    int __gen_e_acsl_and_7;
    __gen_e_acsl_initialized_4 = __e_acsl_initialized((void *)(& p),
                                                      sizeof(char *));
    if (__gen_e_acsl_initialized_4) {
      int __gen_e_acsl_valid_4;
      __gen_e_acsl_valid_4 = __e_acsl_valid((void *)p,sizeof(char),(void *)p,
                                            (void *)(& p));
      __gen_e_acsl_and_5 = __gen_e_acsl_valid_4;
    }
    else __gen_e_acsl_and_5 = 0;
    if (! __gen_e_acsl_and_5) {
      int __gen_e_acsl_initialized_5;
      int __gen_e_acsl_and_6;
      __gen_e_acsl_initialized_5 = __e_acsl_initialized((void *)(& path),
                                                        sizeof(char *));
      if (__gen_e_acsl_initialized_5) {
        int __gen_e_acsl_valid_5;
        __gen_e_acsl_valid_5 = __e_acsl_valid((void *)path,sizeof(char),
                                              (void *)path,(void *)(& path));
        __gen_e_acsl_and_6 = __gen_e_acsl_valid_5;
      }
      else __gen_e_acsl_and_6 = 0;
      __gen_e_acsl_and_7 = ! __gen_e_acsl_and_6;
    }
    else __gen_e_acsl_and_7 = 0;
    __e_acsl_assert(__gen_e_acsl_and_7,"Assertion","main",
                    "!\\valid(p) && !\\valid(path)",
                    "tests/temporal/t_fun_lib.c",30);
  }
  /*@ assert ¬\valid(p) ∧ ¬\valid(path); */ ;
  __retres = 0;
  __e_acsl_delete_block((void *)(& path));
  __e_acsl_delete_block((void *)(& q));
  __e_acsl_delete_block((void *)(& p));
  __e_acsl_memory_clean();
  return __retres;
}


