#include "../../../../lib/stubs.h"

/* Size of input buffers and overflowed buffer; must be able to hold
 * null-terminators in addition to actual data. */
#define MAXPATHLEN BASE_SZ + 2

/* Used in symlinks_bad.c. */
#define MAXSYMLINKS 1

int readlink(const char *path, char *buf, int bufsiz);
char *getcwd(char *buf, size_t size);

