/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
int main(void)
{
  int __retres;
  int x = 0;
  {
    int i = 0;
    while (i < 10) {
      {
        int __gen_e_acsl_and;
        if (0 <= i) __gen_e_acsl_and = i < 10; else __gen_e_acsl_and = 0;
        __e_acsl_assert(__gen_e_acsl_and,"Invariant","main","0 <= i < 10",
                        "tests/constructs/invariant.i",9);
      }
      /*@ invariant 0 ≤ i < 10; */ ;
      x += i;
      __e_acsl_assert(i <= x,"Invariant","main","i <= x",
                      "tests/constructs/invariant.i",11);
      /*@ invariant i ≤ x; */ ;
      i ++;
    }
  }
  __retres = 0;
  return __retres;
}


