/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
void function(int a, int b, int c, int d)
{
  return;
}

int main(void)
{
  int __retres;
  int w = 0;
  int x = 1;
  int y = 2;
  int z = 3;
  function(w,x,y,z);
  __retres = 0;
  return __retres;
}


