/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
char a;
void __e_acsl_globals_init(void)
{
  static char __e_acsl_already_run = 0;
  if (! __e_acsl_already_run) {
    __e_acsl_already_run = 1;
    __e_acsl_store_block((void *)(& a),(size_t)1);
    __e_acsl_full_init((void *)(& a));
  }
  return;
}

void __e_acsl_globals_clean(void)
{
  __e_acsl_delete_block((void *)(& a));
}

int main(void)
{
  int __retres;
  char *b;
  __e_acsl_memory_init((int *)0,(char ***)0,(size_t)8);
  __e_acsl_globals_init();
  __e_acsl_store_block((void *)(& b),(size_t)8);
  goto _LOR;
  _LOR: {
          __e_acsl_full_init((void *)(& b));
          b = & a;
        }
  if (a) goto _LOR;
  {
    int __gen_e_acsl_initialized;
    __gen_e_acsl_initialized = __e_acsl_initialized((void *)b,sizeof(char));
    __e_acsl_assert(__gen_e_acsl_initialized,"Assertion","main",
                    "\\initialized(b)","tests/memory/goto.c",15);
  }
  /*@ assert \initialized(b); */ ;
  __retres = 0;
  __e_acsl_delete_block((void *)(& b));
  __e_acsl_globals_clean();
  __e_acsl_memory_clean();
  return __retres;
}


