/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
void __e_acsl_globals_init(void)
{
  static char __e_acsl_already_run = 0;
  if (! __e_acsl_already_run) {
    __e_acsl_already_run = 1;
    __e_acsl_store_block((void *)(& stdout),(size_t)8);
    __e_acsl_full_init((void *)(& stdout));
    __e_acsl_store_block((void *)(& stdin),(size_t)8);
    __e_acsl_full_init((void *)(& stdin));
    __e_acsl_store_block((void *)(& stderr),(size_t)8);
    __e_acsl_full_init((void *)(& stderr));
  }
  return;
}

void __e_acsl_globals_clean(void)
{
  __e_acsl_delete_block((void *)(& stdout));
  __e_acsl_delete_block((void *)(& stdin));
  __e_acsl_delete_block((void *)(& stderr));
}

int main(void)
{
  int __retres;
  __e_acsl_memory_init((int *)0,(char ***)0,(size_t)8);
  __e_acsl_globals_init();
  {
    int __gen_e_acsl_valid;
    __gen_e_acsl_valid = __e_acsl_valid((void *)stderr,sizeof(FILE),
                                        (void *)stderr,(void *)(& stderr));
    __e_acsl_assert(__gen_e_acsl_valid,"Assertion","main",
                    "\\valid(__fc_stderr)","tests/memory/stdout.c",8);
  }
  /*@ assert \valid(__fc_stderr); */ ;
  {
    int __gen_e_acsl_valid_2;
    __gen_e_acsl_valid_2 = __e_acsl_valid((void *)stdin,sizeof(FILE),
                                          (void *)stdin,(void *)(& stdin));
    __e_acsl_assert(__gen_e_acsl_valid_2,"Assertion","main",
                    "\\valid(__fc_stdin)","tests/memory/stdout.c",9);
  }
  /*@ assert \valid(__fc_stdin); */ ;
  {
    int __gen_e_acsl_valid_3;
    __gen_e_acsl_valid_3 = __e_acsl_valid((void *)stdout,sizeof(FILE),
                                          (void *)stdout,(void *)(& stdout));
    __e_acsl_assert(__gen_e_acsl_valid_3,"Assertion","main",
                    "\\valid(__fc_stdout)","tests/memory/stdout.c",10);
  }
  /*@ assert \valid(__fc_stdout); */ ;
  __retres = 0;
  __e_acsl_globals_clean();
  __e_acsl_memory_clean();
  return __retres;
}


