#include "../../../../lib/stubs.h"

typedef int NSS_STATUS;

/* Size of overflowed buffer. */
#define FSTRING_LEN BASE_SZ /* originally 256 */
typedef char fstring[FSTRING_LEN];

/* Size of input buffer. */
#define INSZ (FSTRING_LEN+2)
