/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
int main(void)
{
  int __retres;
  __e_acsl_memory_init((int *)0,(char ***)0,(size_t)8);
  int *p = (int *)0;
  __e_acsl_temporal_store_nblock((void *)(& p),(void *)0);
  __e_acsl_store_block((void *)(& p),(size_t)8);
  __e_acsl_full_init((void *)(& p));
  int *q = (int *)0;
  __e_acsl_temporal_store_nblock((void *)(& q),(void *)0);
  __e_acsl_store_block((void *)(& q),(size_t)8);
  __e_acsl_full_init((void *)(& q));
  {
    int i = 9;
    __e_acsl_store_block((void *)(& i),(size_t)4);
    __e_acsl_full_init((void *)(& i));
    __e_acsl_full_init((void *)(& p));
    __e_acsl_temporal_store_nblock((void *)(& p),(void *)(& i));
    p = & i;
    __e_acsl_full_init((void *)(& q));
    __e_acsl_temporal_store_nreferent((void *)(& q),(void *)(& p));
    q = p;
    __e_acsl_delete_block((void *)(& i));
  }
  {
    int __gen_e_acsl_initialized;
    int __gen_e_acsl_and;
    __gen_e_acsl_initialized = __e_acsl_initialized((void *)(& p),
                                                    sizeof(int *));
    if (__gen_e_acsl_initialized) {
      int __gen_e_acsl_valid;
      /*@ assert Eva: dangling_pointer: ¬\dangling(&p); */
      __gen_e_acsl_valid = __e_acsl_valid((void *)p,sizeof(int),(void *)p,
                                          (void *)(& p));
      __gen_e_acsl_and = __gen_e_acsl_valid;
    }
    else __gen_e_acsl_and = 0;
    __e_acsl_assert(! __gen_e_acsl_and,"Assertion","main","!\\valid(p)",
                    "tests/temporal/t_scope.c",16);
  }
  /*@ assert ¬\valid(p); */ ;
  {
    int __gen_e_acsl_initialized_2;
    int __gen_e_acsl_and_2;
    __gen_e_acsl_initialized_2 = __e_acsl_initialized((void *)(& q),
                                                      sizeof(int *));
    if (__gen_e_acsl_initialized_2) {
      int __gen_e_acsl_valid_2;
      __gen_e_acsl_valid_2 = __e_acsl_valid((void *)q,sizeof(int),(void *)q,
                                            (void *)(& q));
      __gen_e_acsl_and_2 = __gen_e_acsl_valid_2;
    }
    else __gen_e_acsl_and_2 = 0;
    __e_acsl_assert(! __gen_e_acsl_and_2,"Assertion","main","!\\valid(q)",
                    "tests/temporal/t_scope.c",17);
  }
  /*@ assert ¬\valid(q); */ ;
  {
    int j = 8;
    __e_acsl_store_block((void *)(& j),(size_t)4);
    __e_acsl_full_init((void *)(& j));
    __e_acsl_full_init((void *)(& p));
    __e_acsl_temporal_store_nblock((void *)(& p),(void *)(& j));
    p = & j;
    {
      int __gen_e_acsl_initialized_3;
      int __gen_e_acsl_and_3;
      __gen_e_acsl_initialized_3 = __e_acsl_initialized((void *)(& p),
                                                        sizeof(int *));
      if (__gen_e_acsl_initialized_3) {
        int __gen_e_acsl_valid_3;
        __gen_e_acsl_valid_3 = __e_acsl_valid((void *)p,sizeof(int),
                                              (void *)p,(void *)(& p));
        __gen_e_acsl_and_3 = __gen_e_acsl_valid_3;
      }
      else __gen_e_acsl_and_3 = 0;
      __e_acsl_assert(__gen_e_acsl_and_3,"Assertion","main","\\valid(p)",
                      "tests/temporal/t_scope.c",22);
    }
    /*@ assert \valid(p); */ ;
    __e_acsl_initialize((void *)p,sizeof(int));
    *p = 1;
    {
      int __gen_e_acsl_initialized_4;
      int __gen_e_acsl_and_4;
      __gen_e_acsl_initialized_4 = __e_acsl_initialized((void *)(& q),
                                                        sizeof(int *));
      if (__gen_e_acsl_initialized_4) {
        int __gen_e_acsl_valid_4;
        __gen_e_acsl_valid_4 = __e_acsl_valid((void *)q,sizeof(int),
                                              (void *)q,(void *)(& q));
        __gen_e_acsl_and_4 = __gen_e_acsl_valid_4;
      }
      else __gen_e_acsl_and_4 = 0;
      __e_acsl_assert(! __gen_e_acsl_and_4,"Assertion","main","!\\valid(q)",
                      "tests/temporal/t_scope.c",25);
    }
    /*@ assert ¬\valid(q); */ ;
    {
      int __gen_e_acsl_valid_5;
      __gen_e_acsl_valid_5 = __e_acsl_valid((void *)(& j),sizeof(int),
                                            (void *)(& j),(void *)0);
      __e_acsl_assert(__gen_e_acsl_valid_5,"Assertion","main","\\valid(&j)",
                      "tests/temporal/t_scope.c",26);
    }
    /*@ assert \valid(&j); */ ;
    __e_acsl_delete_block((void *)(& j));
  }
  int len = 3;
  __e_acsl_full_init((void *)(& p));
  __e_acsl_temporal_store_nblock((void *)(& p),(void *)0);
  p = (int *)0;
  __e_acsl_full_init((void *)(& q));
  __e_acsl_temporal_store_nblock((void *)(& q),(void *)0);
  q = (int *)0;
  while (len) {
    int a;
    __e_acsl_store_block((void *)(& a),(size_t)4);
    {
      int __gen_e_acsl_initialized_5;
      int __gen_e_acsl_and_5;
      __gen_e_acsl_initialized_5 = __e_acsl_initialized((void *)(& p),
                                                        sizeof(int *));
      if (__gen_e_acsl_initialized_5) {
        int __gen_e_acsl_valid_6;
        __gen_e_acsl_valid_6 = __e_acsl_valid((void *)p,sizeof(int),
                                              (void *)p,(void *)(& p));
        __gen_e_acsl_and_5 = __gen_e_acsl_valid_6;
      }
      else __gen_e_acsl_and_5 = 0;
      __e_acsl_assert(! __gen_e_acsl_and_5,"Assertion","main","!\\valid(p)",
                      "tests/temporal/t_scope.c",35);
    }
    /*@ assert ¬\valid(p); */ ;
    __e_acsl_full_init((void *)(& q));
    __e_acsl_temporal_store_nblock((void *)(& q),(void *)(& a));
    q = & a;
    __e_acsl_full_init((void *)(& p));
    __e_acsl_temporal_store_nreferent((void *)(& p),(void *)(& q));
    p = q;
    {
      int __gen_e_acsl_initialized_6;
      int __gen_e_acsl_and_6;
      __gen_e_acsl_initialized_6 = __e_acsl_initialized((void *)(& p),
                                                        sizeof(int *));
      if (__gen_e_acsl_initialized_6) {
        int __gen_e_acsl_valid_7;
        __gen_e_acsl_valid_7 = __e_acsl_valid((void *)p,sizeof(int),
                                              (void *)p,(void *)(& p));
        __gen_e_acsl_and_6 = __gen_e_acsl_valid_7;
      }
      else __gen_e_acsl_and_6 = 0;
      __e_acsl_assert(__gen_e_acsl_and_6,"Assertion","main","\\valid(p)",
                      "tests/temporal/t_scope.c",38);
    }
    /*@ assert \valid(p); */ ;
    len --;
    __e_acsl_delete_block((void *)(& a));
  }
  __retres = 0;
  __e_acsl_delete_block((void *)(& q));
  __e_acsl_delete_block((void *)(& p));
  __e_acsl_memory_clean();
  return __retres;
}


