/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
extern int __e_acsl_sound_verdict;

int incr(int x);

/*@ ensures \result ≡ incr(\old(i)); */
int __gen_e_acsl_f(int i);

int f(int i)
{
  int j = i + 1;
  return j;
}

int incr(int x)
{
  int __retres;
  __retres = x + 1;
  return __retres;
}

int main(void)
{
  int __retres;
  int i = __gen_e_acsl_f(2);
  __retres = 0;
  return __retres;
}

/*@ ensures \result ≡ incr(\old(i)); */
int __gen_e_acsl_f(int i)
{
  int __gen_e_acsl_at;
  int __retres;
  __gen_e_acsl_at = i;
  __retres = f(i);
  {
    int __gen_e_acsl_incr_app;
    __gen_e_acsl_incr_app = incr(__gen_e_acsl_at);
    __e_acsl_assert(__retres == __gen_e_acsl_incr_app,"Postcondition","f",
                    "\\result == incr(\\old(i))","tests/special/builtin.i",9);
    return __retres;
  }
}


