/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
char *__gen_e_acsl_literal_string;
char *__gen_e_acsl_literal_string_2;
struct ST {
   char *str ;
   int num ;
};
struct ST _G[2] =
  {{.str = (char *)"Struct_G[0]", .num = 99},
   {.str = (char *)"Struct_G[1]", .num = 147}};
void __e_acsl_globals_init(void)
{
  static char __e_acsl_already_run = 0;
  if (! __e_acsl_already_run) {
    __e_acsl_already_run = 1;
    __gen_e_acsl_literal_string = "Struct_G[1]";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string,
                         sizeof("Struct_G[1]"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string);
    __gen_e_acsl_literal_string_2 = "Struct_G[0]";
    __e_acsl_store_block((void *)__gen_e_acsl_literal_string_2,
                         sizeof("Struct_G[0]"));
    __e_acsl_full_init((void *)__gen_e_acsl_literal_string_2);
    __e_acsl_mark_readonly((void *)__gen_e_acsl_literal_string_2);
    __e_acsl_store_block((void *)(_G),(size_t)32);
    __e_acsl_full_init((void *)(& _G));
  }
  return;
}

void __e_acsl_globals_clean(void)
{
  __e_acsl_delete_block((void *)(_G));
}

int main(int argc, char **argv)
{
  int __retres;
  __e_acsl_memory_init(& argc,& argv,(size_t)8);
  __e_acsl_globals_init();
  {
    int __gen_e_acsl_initialized;
    int __gen_e_acsl_and;
    __gen_e_acsl_initialized = __e_acsl_initialized((void *)(& _G[0].str),
                                                    sizeof(char *));
    if (__gen_e_acsl_initialized) {
      int __gen_e_acsl_valid_read;
      __gen_e_acsl_valid_read = __e_acsl_valid_read((void *)_G[0].str,
                                                    sizeof(char),
                                                    (void *)_G[0].str,
                                                    (void *)(& _G[0].str));
      __gen_e_acsl_and = __gen_e_acsl_valid_read;
    }
    else __gen_e_acsl_and = 0;
    __e_acsl_assert(__gen_e_acsl_and,"Assertion","main",
                    "\\valid_read(_G[0].str)","tests/bts/bts2191.c",22);
  }
  /*@ assert \valid_read(_G[0].str); */ ;
  __retres = 0;
  __e_acsl_globals_clean();
  __e_acsl_memory_clean();
  return __retres;
}


