/* Generated by Frama-C */
#include "stddef.h"
#include "stdint.h"
#include "stdio.h"
int main(int argc, char **argv)
{
  int __retres;
  char *p;
  int *q;
  __e_acsl_memory_init(& argc,& argv,(size_t)8);
  __e_acsl_store_block((void *)(& q),(size_t)8);
  __e_acsl_store_block((void *)(& p),(size_t)8);
  uintptr_t addr = (unsigned long)(& argc);
  __e_acsl_store_block((void *)(& addr),(size_t)8);
  __e_acsl_full_init((void *)(& addr));
  __e_acsl_full_init((void *)(& q));
  __e_acsl_temporal_store_nblock((void *)(& q),(void *)(& argc));
  q = & argc;
  __e_acsl_full_init((void *)(& p));
  __e_acsl_temporal_store_nblock((void *)(& p),(void *)((char *)addr));
  p = (char *)addr;
  __e_acsl_full_init((void *)(& p));
  __e_acsl_temporal_store_nblock((void *)(& p),(void *)0x123456);
  p = (char *)0x123456;
  __e_acsl_full_init((void *)(& p));
  __e_acsl_temporal_store_nreferent((void *)(& p),(void *)(& q));
  p = (char *)q;
  __retres = 0;
  __e_acsl_delete_block((void *)(& argc));
  __e_acsl_delete_block((void *)(& q));
  __e_acsl_delete_block((void *)(& p));
  __e_acsl_delete_block((void *)(& addr));
  __e_acsl_memory_clean();
  return __retres;
}


