/*$************* SCADE_KCG KCG Version 5.1.1 (build i10) **************
** Command :
** l2C        CruiseControl.lus -node CruiseControl
**     -noexp @ALL@
**     -keep_named_var
**     -const
**     -bitwise
**     -loc_ctx
**     -no_copy_mem
**     -debug
** date of generation (MM/DD/YYYY): 07/06/2007 13:30:09
** last modification date for CruiseControl.lus (MM/DD/YYYY): 07/06/2007 
********************************************************************$*/

/* ========= */
/* CONSTANTS */
/* ========= */

extern const Speed ZeroSpeed;
extern const Speed SpeedInc;
extern const Speed SpeedMax;
extern const Speed SpeedMin;
extern const Percent ZeroPercent;
extern const real Kp;
extern const real Ki;
extern const Percent RegThrottleMax;


/*$************* SCADE_KCG KCG Version 5.1.1 (build i10) **************
** End of file CruiseControl_extern.h
** End of generation (MM/DD/YYYY) : 07/06/2007 13:30:09
********************************************************************$*/
