/* run.config
   COMMENT: check a fix of a bug which occured when a [full_init] stmt is
   COMMENT: generated for a labeled statement which a goto points to.
   COMMENT: test both backward and forward gotos.
*/

char a;

int main(void) {
  char *b;
  goto _LOR;
 _LOR: b = &a;
  if (a) goto _LOR; // dead code in order to prevent infinite loop
                    // but still meaningfull in term of code generated by E-ACSL
  /*@ assert \initialized(b); */
  return 0;
}
