/* Generated by Frama-C */
#include "stddef.h"
#include "stdio.h"
int main(void)
{
  int __retres;
  int x = 0;
  x ++;
  __e_acsl_assert(1,"Assertion","main","\\true","tests/constructs/true.i",8);
  /*@ assert \true; */ ;
  __retres = 0;
  return __retres;
}


